/*****************************************************************************

   (c) Copyright Hewlett-Packard Company, 1998.  All rights are
   reserved.  Copying or other reproduction of this program except
   for archival purposes is prohibited without the prior written
   consent of Hewlett-Packard Company.

                   RESTRICTED RIGHTS LEGEND

   Use, duplication, or disclosure by the Government is subject to
   restrictions as set forth in paragraph (b) (3) (B) of the Rights
   in Technical Data and Computer Software clause in DAR 7-104.9(a).

   HEWLETT-PACKARD COMPANY
   Boise, Idaho, USA

******************************************************************************/

/****************************** NOTICE ***************************************

HEWLETT-PACKARD COMPANY MAKES NO WARRANTIES, EXPRESS OR IMPLIED,  
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE
SOFTWARE OR TECHNICAL INFORMATION.   HEWLETT-PACKARD COMPANY DOES
NOT WARRANT, GUARANTEE OR MAKE ANY REPRESENTATIONS REGARDING THE 
USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR TECHNICAL 
INFORMATION IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY, 
CURRENTNESS, OR OTHERWISE.   THE ENTIRE RISK AS TO THE RESULTS AND 
PERFORMANCE OF THE SOFTWARE OR TECHNICAL INFORMATION IS ASSUMED
BY YOU.  The exclusion of implied warranties is not permitted by some 
jurisdictions.  The above exclusion may not apply to you.

IN NO EVENT WILL HEWLETT-PACKARD COMPANY BE LIABLE TO YOU FOR ANY 
CONSEQUENTIAL, INCIDENTAL OR INDIRECT DAMAGES (INCLUDING DAMAGES 
FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
INFORMATION AND THE LIKE)  ARISING OUT OF THE USE OR INABILITY TO USE 
THE SOFTWARE OR TECHNICAL INFORMATION EVEN IF HEWLETT-PACKARD HAS 
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  Because some jurisdictions 
do not allow the exclusion or limitation of liability for consequential
or incidental damages, the above limitations may not apply to you.  
Hewlett-Packard liability to you for actual damages from any cause whatsoever,
 and regardless of the form of the action (whether in contract, tort 
including negligence, product liability or otherwise), will be limited 
to US $50.

Copyright (c) 1998 Hewlett-Packard Company.  All rights reserved.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include "jetlib.h"
#include "xlttlib.h"
#include "debug.h"

#define OUT_BUFFER_MAX_SIZE 4096

/*************************************************************************/
/* Global Variables */

HP_UByte subArray[1] = {HP_eVerticalSubstitution};

HP_UByte HP_UEL[] = "\033%-12345X";

/*  @PJL ENTER LANGUAGE=PCLXL */
HP_UByte HP_Enter_PCLXL[] = "@PJL ENTER LANGUAGE=PCLXL\n";

/*  PXL-XL Comment Array */
HP_UByte HP_PCL_XL_Comment_Array[] =
  "Copyright Hewlett-Packard Company1989-1996";

/*************************************************************************/
/* Function prototypes */

HP_StdFuncPrefix HP_FlushOutBuffer
(
  HP_StreamHandleType pStream,
  HP_UInt32           cookie,
  HP_pUByte           pOutBuffer,
  HP_SInt32           StreamBuffLen
);

void show_usage
(
  char *pgm
);

void reportError
(
  xlttlib_ErrorType error
);

HP_pUByte myAlloc
(
  HP_UInt16 byteCount
);

void myFree
(
  HP_pUByte buffer
);

boolean mySeek
(
  void      *clientDataPtr,
  HP_UInt32  offsetFromTop
);

boolean myRead
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  byteBuffer
);

/*************************************************************************/

void reportError
(
  xlttlib_ErrorType error
)
{
  switch(error)
  {
    case xlttlib_EInvalidMemAccess:
      printf("client: reportError: EInvalidMemAccess\n");
      break;
    case xlttlib_EInvalidCallSequence:
      printf("client: reportError: EInvalidCallSequence\n");
      break;
    case xlttlib_EInvalidFontInfoPointer:
      printf("client: reportError: EInvalidFontInfoPointer\n");
      break;
    case xlttlib_EInvalidFontAccess:
      printf("client: reportError: EInvalidFontAccess\n");
      break;
    case xlttlib_EInvalidFont:
      printf("client: reportError: EInvalidFont\n");
      break;
    case xlttlib_EOutOfFontHandles:
      printf("client: reportError: EOutOfFontHandles\n");
      break;
    case xlttlib_EInvalidFontHandle:
      printf("client: reportError: EInvalidFontHandle\n");
      break;
    case xlttlib_EInvalidCharRange:
      printf("client: reportError: EInvalidCharRange\n");
      break;
    case xlttlib_EInvalidWidthsPtr:
      printf("client: reportError: EInvalidWidthsPtr\n");
      break;
    case xlttlib_ESeekError:
      printf("client: reportError: ESeekError\n");
      break;
    case xlttlib_EReadError:
      printf("client: reportError: EReadError\n");
      break;
    case xlttlib_EMemAllocError:
      printf("client: reportError: EMemAllocError\n");
      break;
    case xlttlib_EUnknown:
    default:
      printf("client: Unknown error (0x%X)\n", error);
      break;
  }
}

/*************************************************************************/

HP_pUByte myAlloc
(
  HP_UInt16 byteCount
)
{
  HP_pUByte buffer = NULL;

  buffer = (HP_pUByte)malloc(byteCount);
  if(!buffer)
  {
    DBG1("client: myAlloc: malloc failed\n");
  }

  DBG2("client: allocation request filled at 0x%X\n", buffer);
  return(buffer);
}

/*************************************************************************/

void myFree
(
    HP_pUByte buffer
)
{
    DBG2("client: free request filled at 0x%X\n", buffer);
    free(buffer);
}

/*************************************************************************/

boolean mySeek
(
  void      *clientDataPtr,
  HP_UInt32  offsetFromTop
)
{
  int ret;

  ret = fseek((FILE *)clientDataPtr, offsetFromTop, SEEK_SET);
  if(ret != 0)
  {
    DBG1("client: Seek Error\n");
    return(FALSE);
  }

  return(TRUE);
}

/*************************************************************************/

boolean myRead
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  byteBuffer
)
{
  size_t ret;

  ret = fread((void *)byteBuffer, 1, byteCount, (FILE *)clientDataPtr);
  if(ret != byteCount)
  {
    DBG1("client: Read Error\n");
    return(FALSE);
  }

  return(TRUE);
}

/*************************************************************************/

void show_usage
(
  char *pgm
)
{
  fprintf(stderr,"%s V1.0\n", pgm);
  fprintf(stderr,
    "(c) Copyright Hewlett-Packard Company, 1998. All rights reserved.\n\n");
  fprintf(stderr,
    "USAGE: %s OutputFileName FontName FontFileName\n\n", pgm);
}

/*************************************************************************/

void PrintCourierRoman8
(
  HP_StreamHandleType pStream,
  HP_Real32           charSize,
  HP_SInt16           xPos,
  HP_SInt16           yPos,
  HP_pUByte           string
)
{
  HP_UInt32  i;
  HP_SInt16 *escapements;
  HP_UInt16  length;

  length = (HP_UInt16)strlen((const char *)string);

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * length);
  if(!escapements)
  {
    fprintf(stderr, "demo: warning: malloc failed");
    return;
  }

  for(i=0;i<length;i++)
  {
    escapements[(HP_UInt16)i] = (HP_SInt16)((charSize + 0.5) / 2);
  }

  HP_PushGS_1(pStream);
  HP_SetColorSpace_1(pStream, HP_eGray);
  HP_SetPenSource_Gray1(pStream, 0.0);
  HP_SetBrushSource_Gray1(pStream, 0.0);
  HP_SetFont_1(pStream, (HP_pUByte)"Courier         ", 16, charSize, 277);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_Text_1(pStream, string, length, escapements, length);  
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void Border
(
  HP_StreamHandleType     pStream
)
{
  HP_PushGS_1(pStream);
  HP_SetColorSpace_1(pStream, HP_eGray);
  HP_SetPenSource_Gray1(pStream, 0.0);
  HP_SetBrushSource_Null1(pStream);
  HP_Rectangle_1(pStream, 100, 100, 5000, 6500);
  HP_PopGS_1(pStream);
}

/*************************************************************************/

void PlainText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
      free(escapements);
      return;
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_Text_4(pStream, string, stringLength, escapements, stringLength);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void PseudoItalicText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
      free(escapements);
      return;
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_SetCharShear_2(pStream, 0.31623, 0.0);
  HP_Text_4(pStream, string, stringLength, escapements, stringLength);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void PseudoBoldText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
      free(escapements);
      return;
    }
    escapements[(HP_UInt16)i] *= 1.015;
  }  

  HP_PushGS_1(pStream);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_SetCharBoldValue_1(pStream, 0.015);
  HP_Text_4(pStream, string, stringLength, escapements, stringLength);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void VerticalRotatedText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;
  HP_UByte   subModeArray[1] = {HP_eVerticalSubstitution};

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, TRUE,
      TRUE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
      free(escapements);
      return;
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetCharSubMode_1(pStream, subModeArray, 1);
  HP_SetCharAttribute_1(pStream, HP_eWriteVertical);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_Text_4(pStream, string, stringLength, escapements, stringLength);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void GrayText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetColorSpace_1(pStream, HP_eGray);
  HP_SetPenSource_Null1(pStream);
  HP_SetBrushSource_Gray1(pStream, 0.75);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_TextPath_4(pStream, string, stringLength, escapements, stringLength);
  HP_PaintPath_1(pStream);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void PatternText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;
  HP_UByte   pallette[2] = {0, 255};
  HP_UByte   pattern[32] = {0x0f, 00,   00, 00, 0x87, 00,   00, 00, 0xc3, 00,
                              00, 00, 0xe1, 00,   00, 00, 0xf0, 00,   00, 00,
                            0x78, 00,   00, 00, 0x3c, 00,   00, 00, 0x1e, 00,
			      00, 00};

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetColorSpace_2(pStream, HP_eGray, HP_e8Bit, pallette, 2);
  HP_BeginRastPattern_1(pStream, HP_eIndexedPixel, HP_e1Bit, 8, 8, 32, 32, 1,
    HP_eTempPattern);
  HP_ReadRastPattern_1(pStream, 0, 8, HP_eNoCompression);
  HP_EmbeddedDataPrefix(pStream, 32);
  HP_RawHugeUByteArray(pStream, (HP_pCharHuge)pattern, 32);
  HP_EndRastPattern_1(pStream);
  HP_SetBrushSource_Patt1(pStream, 1);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_TextPath_4(pStream, string, stringLength, escapements, stringLength);
  HP_PaintPath_1(pStream);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void ShadowText
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  HP_pUByte               fontName,
  HP_Real32               charSize,
  HP_UInt16               mapping,
  HP_SInt16               xPos,
  HP_SInt16               yPos,
  HP_UInt16              *string,
  HP_UInt16               stringLength
)
{
  boolean    ret;
  HP_UInt32  i;
  HP_SInt16 *escapements;

  escapements = (HP_SInt16 *)malloc(sizeof(HP_SInt16) * stringLength);
  if(!escapements)
  {
    fprintf(stderr, "demo: malloc failed\n");
    return;
  }

  for(i=0;i<stringLength;i++)
  {
    ret = xlttlib_GetLinearScaledCharWidths(fontHandle, charSize, FALSE,
      FALSE, string[(HP_UInt16)i], string[(HP_UInt16)i],
      &escapements[(HP_UInt16)i]);
    if(!ret)
    {
      fprintf(stderr, "demo: xlttlib_GetLinearScaledCharWidths failed\n");
    }
  }  

  HP_PushGS_1(pStream);
  HP_SetColorSpace_1(pStream, HP_eGray);
  HP_SetFont_1(pStream, fontName, (HP_UInt16)strlen((const char *)fontName),
    charSize, mapping);
  HP_SetCursor_1(pStream, xPos+10, yPos+10);
  HP_Text_4(pStream, string, stringLength, escapements, stringLength);
  HP_SetBrushSource_Gray1(pStream, 1.0);
  HP_SetCursor_1(pStream, xPos, yPos);
  HP_TextPath_4(pStream, string, stringLength, escapements, stringLength);
  HP_PaintPath_1(pStream);
  HP_PopGS_1(pStream);

  free(escapements);
}

/*************************************************************************/

void DemoTradChinese
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 tradChineseString[20] = {   'a',    'b',    'A',    'B',
                                        '(',    ')',    ',',    '@',
                                     0xa2e9, 0xa2ea, 0xa2cf, 0xa2d0,
                                     0xa15d, 0xa15e, 0xa14e, 0xa249,
                                     0xa5a1, 0xa5a2, 0xa5a3, 0xa5a4};
  HP_UInt16 tradChineseGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Traditional Chinese Demo");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain 1-byte and 2-byte Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200, 900,
    tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic 1-byte and 2-byte Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596,
    200, 1500, tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold 1-byte and 2-byte Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200,
    2100, tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 2500,
    (HP_pUByte)"Vertical-Rotated 1-byte and 2-byte Text:");
  VerticalRotatedText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596,
    200, 2700, tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray 1-byte and 2-byte Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200, 3200,
    tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled 1-byte and 2-byte Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200, 3700,
    tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed 1-byte and 2-byte Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200, 4200,
    tradChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 596, 200, 4700,
    tradChineseGalley, 1);
}

/*************************************************************************/

void DemoSimpChinese
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 simpChineseString[20] = {   'a',    'b',    'A',    'B',
                                        '(',    ')',    ',',    '@',
                                     0xa3e1, 0xa3e2, 0xa3c1, 0xa3c2,
                                     0xa3a8, 0xa3a9, 0xa3ac, 0xa3c0,
                                     0xb0a1, 0xb0a2, 0xb0a3, 0xb0a4};
  HP_UInt16 simpChineseGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Simplified Chinese Demo");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain 1-byte and 2-byte Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200, 900,
    simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic 1-byte and 2-byte Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643,
    200, 1500, simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold 1-byte and 2-byte Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200,
    2100, simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 2500,
    (HP_pUByte)"Vertical-Rotated 1-byte and 2-byte Text:");
  VerticalRotatedText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643,
    200, 2700, simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray 1-byte and 2-byte Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200, 3200,
    simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled 1-byte and 2-byte Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200, 3700,
    simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed 1-byte and 2-byte Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200, 4200,
    simpChineseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 643, 200, 4700,
    simpChineseGalley, 1);
}

/*************************************************************************/

void DemoJapanese
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 japaneseString[20] = {   'a',    'A',    0x00b1, 0x00b2,
                                     '(',    ')',    ',',    '@',
                                  0x8281, 0x8260, 0x8341, 0x8343,
                                  0x8169, 0x816a, 0x8143, 0x8197,
                                  0x889f, 0x88a0, 0x88a1, 0x88a2};
  HP_UInt16 halfWidthGalley[1] = {0x0000};
  HP_UInt16 fullWidthGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Japanese Demo");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain 1-byte and 2-byte Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 900,
    japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic 1-byte and 2-byte Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619,
    200, 1500, japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold 1-byte and 2-byte Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200,
    2100, japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 2500,
    (HP_pUByte)"Vertical-Rotated 1-byte and 2-byte Text:");
  VerticalRotatedText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619,
    200, 2700, japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray 1-byte and 2-byte Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 3200,
    japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled 1-byte and 2-byte Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 3700,
    japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed 1-byte and 2-byte Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 4200,
    japaneseString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default 1-byte character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 4700,
    halfWidthGalley, 1);
  PrintCourierRoman8(pStream, 200.0, 200, 5000,
    (HP_pUByte)"Default 2-byte character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 619, 200, 5200,
    fullWidthGalley, 1);
}

/*************************************************************************/

void DemoKorean
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 koreanString[20] = {   'a',    'b',    'A',    'B',
                                   '(',    ')',    ',',    '@',
                                0xa3e1, 0xa3e2, 0xa3c1, 0xa3c2,
                                0xa3a8, 0xa3a9, 0xa3ac, 0xa3c0,
                                0xb0a1, 0xb0a2, 0xb0a3, 0xb0a4};
  HP_UInt16 koreanGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Korean (Wansung) Demo");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain 1-byte and 2-byte Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200, 900,
    koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic 1-byte and 2-byte Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616,
    200, 1500, koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold 1-byte and 2-byte Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200,
    2100, koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 2500,
    (HP_pUByte)"Vertical-Rotated 1-byte and 2-byte Text:");
  VerticalRotatedText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616,
    200, 2700, koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray 1-byte and 2-byte Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200, 3200,
    koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled 1-byte and 2-byte Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200, 3700,
    koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed 1-byte and 2-byte Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200, 4200,
    koreanString, 20);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 616, 200, 4700,
    koreanGalley, 1);
}

/*************************************************************************/

void DemoSymbol
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 symbolString[6] = {0xf061, 0xf062, 0xf063,  /* "abc" */
                               0xf041, 0xf042, 0xf043}; /* "ABC" */
  HP_UInt16 symbolGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Symbol Demo using \"abcABC\"");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200, 900,
    symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621,
    200, 1500, symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200,
    2100, symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200, 3200,
    symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200, 3700,
    symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200, 4200,
    symbolString, 6);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 621, 200, 4700,
    symbolGalley, 1);
}

/*************************************************************************/

void DemoUnicode
(
  HP_StreamHandleType     pStream,
  xlttlib_FontHandleType  fontHandle,
  char                   *fontName
)
{
  HP_UInt16 unicodeString[17] = {   'a',    'b',    'A',    'B',
                                    '(',    ')',    ',',    '@',
                                 0xff41, 0xff42, 0xff21, 0xff22,
                                 0xff08, 0xff09, 0xff0c, 0xff20,
                                 0x51fa}; /* Unified CJK */
  HP_UInt16 unicodeGalley[1] = {0xFFFF};

  PrintCourierRoman8(pStream, 200.0, 200, 300,
    (HP_pUByte)"Unicode Demo");
  PrintCourierRoman8(pStream, 200.0, 200, 700,
    (HP_pUByte)"Plain Text:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200, 900,
    unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 1300,
    (HP_pUByte)"Pseudo-Italic Text:");
  PseudoItalicText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590,
    200, 1500, unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 1900,
    (HP_pUByte)"Pseudo-Bold Text:");
  PseudoBoldText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200,
    2100, unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 2500,
    (HP_pUByte)"Vertical-Rotated 1-byte and 2-byte Text:");
  VerticalRotatedText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590,
    200, 2700, unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 3000,
    (HP_pUByte)"Gray Text (no border):");
  GrayText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200, 3200,
    unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 3500,
    (HP_pUByte)"Pattern Filled Text (border):");
  PatternText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200, 3700,
    unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 4000,
    (HP_pUByte)"Shadowed Text:");
  ShadowText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200, 4200,
    unicodeString, 17);
  PrintCourierRoman8(pStream, 200.0, 200, 4500,
    (HP_pUByte)"Default character for this font:");
  PlainText(pStream, fontHandle, (HP_pUByte)fontName, 200.0, 590, 200, 4700,
    unicodeGalley, 1);
}

/*************************************************************************/

void main(
  int   argc,
  char *argv[]
)
{
  HP_StreamHandleType     pStream;
  FILE                   *fontFile = NULL;
  char                   *fontFileName = NULL;
  FILE                   *outFile = NULL;
  char                   *outFileName = NULL;
  xlttlib_MemAccessType   memAccess;
  xlttlib_FontAccessType  fontAccess;
  xlttlib_FontInfoType    fontInfo;
  boolean                 ret;
  char                   *fontName = NULL;

  if(argc != 4)
  {
    show_usage(argv[0]);
    exit(1);
  }

  outFileName = argv[1];
  fontName = argv[2];
  fontFileName = argv[3];

  fontFile = fopen(fontFileName, "rb");
  if(fontFile == NULL)
  {
    fprintf(stderr, "client: Unable to open %s for reading\n",
      fontFileName);
    exit(1);
  }

  outFile = fopen(outFileName, "wb");
  if(outFile == NULL)
  {
    fprintf(stderr, "client: Unable to create %s for writing\n", outFileName);
    exit(1);
  }

  memAccess.memAlloc = myAlloc;
  memAccess.memFree  = myFree;

  ret = xlttlib_Initialize(memAccess);
  if(!ret)
  {
    fprintf(stderr, "client: xlttlib_Initialize failed\n");
    reportError(xlttlib_GetError());
    fclose(fontFile);
    fclose(outFile);
    exit(1);
  }
  fontAccess.seek          = mySeek;
  fontAccess.read          = myRead;
  fontAccess.clientDataPtr = (void *)fontFile;

  ret = xlttlib_RegisterFont(fontAccess, &fontInfo);
  if(!ret)
  {
    fprintf(stderr, "client: xlttlib_RegisterFont failed\n");
    reportError(xlttlib_GetError());
    xlttlib_ShutDown();
    fclose(fontFile);
    fclose(outFile);
    exit(1);
  }
  printf("fontHandle = %u\n", fontInfo.fontHandle);
  printf("%s mapping = %u\n", fontFileName, fontInfo.mapping);

  pStream = HP_NewStream(OUT_BUFFER_MAX_SIZE, HP_FlushOutBuffer,
    (unsigned long)outFile);
  if(!pStream)
  {
    fprintf(stderr,"Error: Stream not initialized.\n");
    xlttlib_ShutDown();
    fclose(fontFile);
    fclose(outFile);
    exit(1);
  }

  HP_RawUByteArray(pStream, HP_UEL, 9);

  HP_RawUByteArray(pStream, HP_Enter_PCLXL, 26);

  HP_AddStreamHeader_1(pStream, 2, 0, HP_PCL_XL_Comment_Array, 43);

  HP_BeginSession_2(pStream, 600, 600, HP_eInch, HP_eBackChAndErrPage);

  HP_OpenDataSource_1(pStream, HP_eDefaultDataSource, HP_eBinaryLowByteFirst);

  ret = xlttlib_DownloadEntireFont(pStream, fontInfo.fontHandle,
    (HP_pUByte)fontName);
  if(!ret)
  {
    fprintf(stderr, "client: xlttlib_DownloadEntireFont failed\n");
    reportError(xlttlib_GetError());
    xlttlib_ShutDown();
    fclose(fontFile);
    fclose(outFile);
    /* Flush UEL from the buffer out to the file */
    pStream->pFlushOutBuffer(pStream, pStream->cookie,
      pStream->HP_OutBuffer, pStream->HP_CurrentBufferLen);
    HP_FinishStream(pStream);
    exit(1);
  }

  HP_BeginPage_1(pStream, HP_ePortraitOrientation, HP_eLetterPaper);

  Border(pStream);

  switch(fontInfo.mapping)
  {
    case xlttlib_ESymbolMapping:
      DemoSymbol(pStream, fontInfo.fontHandle, fontName);
      break;
    case xlttlib_EUnicodeMapping:
      DemoUnicode(pStream, fontInfo.fontHandle, fontName);
      break;
    case xlttlib_EJapaneseMapping:
      DemoJapanese(pStream, fontInfo.fontHandle, fontName);
      break;
    case xlttlib_ESimpChineseMapping:
      DemoSimpChinese(pStream, fontInfo.fontHandle, fontName);
      break;
    case xlttlib_ETradChineseMapping:
      DemoTradChinese(pStream, fontInfo.fontHandle, fontName);
      break;
    case xlttlib_EKoreanWansungMapping:
      DemoKorean(pStream, fontInfo.fontHandle, fontName);
      break;
    default:
      fprintf(stderr, "client: unknown mapping\n");
      break;
  }

  HP_EndPage_2(pStream, 1);

  HP_CloseDataSource_1(pStream);

  HP_EndSession_1(pStream);

  HP_RawUByteArray(pStream, HP_UEL, 9);

  /* Flush UEL from the buffer out to the file */
  pStream->pFlushOutBuffer(pStream, pStream->cookie, pStream->HP_OutBuffer,
    pStream->HP_CurrentBufferLen);

  /* Free the memory allocated for the buffer */
  HP_FinishStream(pStream);

  fclose(outFile);

  if(fontFile)
  {
    xlttlib_ShutDown();
    fclose(fontFile);
  }

  printf("\nDONE\n\n");
  exit(0);
}

/*************************************************************************/

HP_StdFuncPrefix HP_FlushOutBuffer
(
  HP_StreamHandleType pStream,
  unsigned long       cookie,
  HP_pUByte           pOutBuffer,
  HP_SInt32           StreamBuffLen
)
{
  HP_SInt32  index;
  FILE      *outFile = (FILE *)cookie;

  index = 0;
  while(index != StreamBuffLen)
  {
    fputc(pOutBuffer[(HP_UInt16)index], outFile);
    index++;
  }

  pStream->HP_CurrentBufferLen = 0;  /* set outbuffer to empty */
}
