/*****************************************************************************

   (c) Copyright Hewlett-Packard Company, 1998.  All rights are
   reserved.  Copying or other reproduction of this program except
   for archival purposes is prohibited without the prior written
   consent of Hewlett-Packard Company.

                   RESTRICTED RIGHTS LEGEND

   Use, duplication, or disclosure by the Government is subject to
   restrictions as set forth in paragraph (b) (3) (B) of the Rights
   in Technical Data and Computer Software clause in DAR 7-104.9(a).

   HEWLETT-PACKARD COMPANY
   Boise, Idaho, USA

******************************************************************************/

/****************************** NOTICE ***************************************

HEWLETT-PACKARD COMPANY MAKES NO WARRANTIES, EXPRESS OR IMPLIED,  
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE
SOFTWARE OR TECHNICAL INFORMATION.   HEWLETT-PACKARD COMPANY DOES
NOT WARRANT, GUARANTEE OR MAKE ANY REPRESENTATIONS REGARDING THE 
USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR TECHNICAL 
INFORMATION IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY, 
CURRENTNESS, OR OTHERWISE.   THE ENTIRE RISK AS TO THE RESULTS AND 
PERFORMANCE OF THE SOFTWARE OR TECHNICAL INFORMATION IS ASSUMED
BY YOU.  The exclusion of implied warranties is not permitted by some 
jurisdictions.  The above exclusion may not apply to you.

IN NO EVENT WILL HEWLETT-PACKARD COMPANY BE LIABLE TO YOU FOR ANY 
CONSEQUENTIAL, INCIDENTAL OR INDIRECT DAMAGES (INCLUDING DAMAGES 
FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
INFORMATION AND THE LIKE)  ARISING OUT OF THE USE OR INABILITY TO USE 
THE SOFTWARE OR TECHNICAL INFORMATION EVEN IF HEWLETT-PACKARD HAS 
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  Because some jurisdictions 
do not allow the exclusion or limitation of liability for consequential
or incidental damages, the above limitations may not apply to you.  
Hewlett-Packard liability to you for actual damages from any cause whatsoever,
 and regardless of the form of the action (whether in contract, tort 
including negligence, product liability or otherwise), will be limited 
to US $50.

Copyright (c) 1998 Hewlett-Packard Company.  All rights reserved.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "jetlib.h"
#include "xlttlib.h"
#include "debug.h"

#define OUT_BUFFER_MAX_SIZE 4096

/*************************************************************************/
/* Global Variables */

HP_UByte subArray[1] = {HP_eVerticalSubstitution};

HP_UByte HP_UEL[] = "\033%-12345X";

/*  @PJL ENTER LANGUAGE=PCLXL */
HP_UByte HP_Enter_PCLXL[] = "@PJL ENTER LANGUAGE=PCLXL\n";

/*  PXL-XL Comment Array */
HP_UByte HP_PCL_XL_Comment_Array[] =
  "Copyright Hewlett-Packard Company1989-1996";

/*************************************************************************/
/* Function prototypes */

HP_StdFuncPrefix HP_FlushOutBuffer
(
  HP_StreamHandleType pStream,
  HP_UInt32           cookie,
  HP_pUByte           pOutBuffer,
  HP_SInt32           StreamBuffLen
);

void drawGrid
(
  HP_StreamHandleType pStream
);

void drawLabels
(
  HP_StreamHandleType pStream,
  HP_UByte            block
);

void drawHeader
(
  HP_StreamHandleType pStream,
  HP_pUByte fontName
);

void drawFooter
(
  HP_StreamHandleType pStream,
  HP_UByte            block,
  HP_UInt16           mapping
);

void beginPage
(
  HP_StreamHandleType pStream
);

void setNormalOriginAndScale
(
  HP_StreamHandleType pStream
);

void setFourUpOriginAndScale
(
  HP_StreamHandleType pStream,
  HP_UByte block
);

void endPage
(
  HP_StreamHandleType pStream
);

void drawPages
(
  HP_StreamHandleType pStream,
  HP_pUByte           fontName,
  HP_UInt16           mapping,
  HP_UByte            firstBlock,
  HP_UByte            lastBlock,
  boolean             fourUp,
  boolean             vSub,
  boolean             vRot
);

void show_usage
(
  char *pgm
);

void reportError
(
  xlttlib_ErrorType error
);

HP_pUByte myAlloc
(
  HP_UInt16 byteCount
);

void myFree
(
  HP_pUByte buffer
);

boolean mySeek
(
  void      *clientDataPtr,
  HP_UInt32  offsetFromTop
);

boolean myRead
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  byteBuffer
);

boolean myStreamWrite
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  dataPtr
);

/*************************************************************************/

void reportError
(
  xlttlib_ErrorType error
)
{
  switch(error)
  {
    case xlttlib_EInvalidMemAccess:
      printf("client: reportError: EInvalidMemAccess\n");
      break;
    case xlttlib_EInvalidCallSequence:
      printf("client: reportError: EInvalidCallSequence\n");
      break;
    case xlttlib_EInvalidFontInfoPointer:
      printf("client: reportError: EInvalidFontInfoPointer\n");
      break;
    case xlttlib_EInvalidFontAccess:
      printf("client: reportError: EInvalidFontAccess\n");
      break;
    case xlttlib_EInvalidFont:
      printf("client: reportError: EInvalidFont\n");
      break;
    case xlttlib_EOutOfFontHandles:
      printf("client: reportError: EOutOfFontHandles\n");
      break;
    case xlttlib_EInvalidFontHandle:
      printf("client: reportError: EInvalidFontHandle\n");
      break;
    case xlttlib_EInvalidCharRange:
      printf("client: reportError: EInvalidCharRange\n");
      break;
    case xlttlib_EInvalidWidthsPtr:
      printf("client: reportError: EInvalidWidthsPtr\n");
      break;
    case xlttlib_ESeekError:
      printf("client: reportError: ESeekError\n");
      break;
    case xlttlib_EReadError:
      printf("client: reportError: EReadError\n");
      break;
    case xlttlib_EMemAllocError:
      printf("client: reportError: EMemAllocError\n");
      break;
    case xlttlib_EUnknown:
    default:
      printf("client: Unknown error (0x%X)\n", error);
      break;
  }
}

/*************************************************************************/

HP_pUByte myAlloc
(
  HP_UInt16 byteCount
)
{
  HP_pUByte buffer = NULL;

  buffer = (HP_pUByte)malloc(byteCount);
  if(!buffer)
  {
    DBG1("client: myAlloc: malloc failed\n");
  }

  DBG2("client: allocation request filled at 0x%X\n", buffer);
  return(buffer);
}

/*************************************************************************/

void myFree
(
    HP_pUByte buffer
)
{
    DBG2("client: free request filled at 0x%X\n", buffer);
    free((void *)buffer);
}

/*************************************************************************/

boolean mySeek
(
  void      *clientDataPtr,
  HP_UInt32  offsetFromTop
)
{
  int ret;

  ret = fseek((FILE *)clientDataPtr, offsetFromTop, SEEK_SET);
  if(ret != 0)
  {
    DBG1("client: Seek Error\n");
    return(FALSE);
  }

  return(TRUE);
}

/*************************************************************************/

boolean myRead
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  byteBuffer
)
{
  size_t ret;

  ret = fread((void *)byteBuffer, 1, byteCount, (FILE *)clientDataPtr);
  if(ret != byteCount)
  {
    DBG1("client: Read Error\n");
    return(FALSE);
  }

  return(TRUE);
}

/*************************************************************************/

boolean myStreamWrite
(
  void      *clientDataPtr,
  HP_UInt16  byteCount,
  HP_pUByte  dataPtr
)
{
  size_t ret;

  ret = fwrite((const void *)dataPtr, 1, byteCount, (FILE *)clientDataPtr);
  if(ret != byteCount)
  {
    DBG1("client: Write Error\n");
    return(FALSE);
  }

  return(TRUE);
}

/*************************************************************************/

void drawGrid
(
  HP_StreamHandleType pStream
)
{
  HP_SInt16 xEscapements[32] = {90, 127, 90, 127, 90, 127, 90, 127, 90, 127,
                                90, 127, 90, 127, 90, 127, 90, 127, 90, 127,
                                90, 127, 90, 127, 90, 127, 90, 127, 90, 127,
                                90, 127};

  HP_Rectangle_1(pStream,  893, 1186,  901, 5206);
  HP_Rectangle_1(pStream, 1110, 1194, 1118, 5206);
  HP_Rectangle_1(pStream, 1327, 1194, 1335, 5206);
  HP_Rectangle_1(pStream, 1544, 1194, 1552, 5206);
  HP_Rectangle_1(pStream, 1761, 1194, 1769, 5206);
  HP_Rectangle_1(pStream, 1978, 1194, 1986, 5206);
  HP_Rectangle_1(pStream, 2195, 1194, 2203, 5206);
  HP_Rectangle_1(pStream, 2412, 1194, 2420, 5206);
  HP_Rectangle_1(pStream, 2629, 1194, 2637, 5206);
  HP_Rectangle_1(pStream, 2846, 1194, 2854, 5206);
  HP_Rectangle_1(pStream, 3063, 1194, 3071, 5206);
  HP_Rectangle_1(pStream, 3280, 1194, 3288, 5206);
  HP_Rectangle_1(pStream, 3497, 1194, 3505, 5206);
  HP_Rectangle_1(pStream, 3714, 1194, 3722, 5206);
  HP_Rectangle_1(pStream, 3931, 1194, 3939, 5206);
  HP_Rectangle_1(pStream, 4148, 1194, 4156, 5206);
  HP_Rectangle_1(pStream, 4365, 1194, 4373, 5206);
  HP_Rectangle_1(pStream,  518, 1422,  526, 5206);
  HP_Rectangle_1(pStream,  901, 1186, 4373, 1194);
  HP_Rectangle_1(pStream,  526, 1422, 4373, 1430);
  HP_Rectangle_1(pStream,  526, 1658, 4373, 1666);
  HP_Rectangle_1(pStream,  526, 1894, 4373, 1902);
  HP_Rectangle_1(pStream,  526, 2130, 4373, 2138);
  HP_Rectangle_1(pStream,  526, 2366, 4373, 2374);
  HP_Rectangle_1(pStream,  526, 2602, 4373, 2610);
  HP_Rectangle_1(pStream,  526, 2838, 4373, 2846);
  HP_Rectangle_1(pStream,  526, 3074, 4373, 3082);
  HP_Rectangle_1(pStream,  526, 3310, 4373, 3318);
  HP_Rectangle_1(pStream,  526, 3546, 4373, 3554);
  HP_Rectangle_1(pStream,  526, 3782, 4373, 3790);
  HP_Rectangle_1(pStream,  526, 4018, 4373, 4026);
  HP_Rectangle_1(pStream,  526, 4254, 4373, 4262);
  HP_Rectangle_1(pStream,  526, 4490, 4373, 4498);
  HP_Rectangle_1(pStream,  526, 4726, 4373, 4734);
  HP_Rectangle_1(pStream,  526, 4962, 4373, 4970);
  HP_Rectangle_1(pStream,  526, 5198, 4373, 5206);

  HP_SetFont_1(pStream, (HP_pUByte)"Courier         ", 16, 150, 277);

  HP_SetCursor_1(pStream, 919, 1346);

  HP_Text_1(pStream, (HP_pUByte)"_0_1_2_3_4_5_6_7_8_9_A_B_C_D_E_F", 32,
    (HP_pSInt16)xEscapements, 32);
}

/*************************************************************************/

void drawLabels
(
  HP_StreamHandleType pStream,
  HP_UByte            block
)
{
  HP_UByte  line;
  char      characters[5];
  HP_SInt16 yOffset;
  HP_SInt16 escapements[4] = {90, 90, 90, 90};

  HP_SetFont_1(pStream, (HP_pUByte)"Courier         ", 16, 150, 277);

  for(line=0;line<16;line++)
  {
    yOffset = 1582 + (line * 236);

    HP_SetCursor_1(pStream, 533, yOffset);
    sprintf(characters, "%03X_", ((block * 16) + line));
    HP_Text_1(pStream, (HP_pUByte)characters, 4, (HP_pSInt16)escapements, 4);
  }
}

/*************************************************************************/

void drawHeader
(
  HP_StreamHandleType pStream,
  HP_pUByte           fontName
)
{
  char      header[120];
  HP_SInt16 escapements[120];
  HP_UInt16 len;
  int       i;

  if(strlen((const char *)fontName) > 100)
  {
    fprintf(stderr, "\nError: FontName too long\n\n");
    exit(1);
  }

  HP_SetFont_1(pStream, (HP_pUByte)"Courier       Bd", 16, 200, 277);
  HP_SetCursor_1(pStream, 533, 1050);
  sprintf(header, "Font: '%s'", fontName);
  len = (HP_UInt16)strlen((const char *)header);
  for(i=0;i<len;i++)
  {
    escapements[i] = 120;
  }
  HP_Text_1(pStream, (HP_pUByte)header, len, (HP_pSInt16)escapements, len);
}

/*************************************************************************/

void drawFooter
(
  HP_StreamHandleType pStream,
  HP_UByte            block,
  HP_UInt16           mapping
)
{
  char      footer[31];
  HP_SInt16 escapements[30] = {120, 120, 120, 120, 120, 120, 120, 120,
                               120, 120, 120, 120, 120, 120, 120, 120,
                               120, 120, 120, 120, 120, 120, 120, 120,
                               120, 120, 120, 120, 120, 120};

  HP_SetFont_1(pStream, (HP_pUByte)"Courier       Bd", 16, 200, 277);
  HP_SetCursor_1(pStream, 533, 5500);
  sprintf(footer,
    "Mapping %3d, Block %3d (0x%02X)", mapping, block, block);
  HP_Text_1(pStream, (HP_pUByte)footer, 30, (HP_pSInt16)escapements, 30);
}

/*************************************************************************/

void beginPage
(
  HP_StreamHandleType pStream
)
{
  HP_BeginPage_1(pStream, HP_ePortraitOrientation, HP_eLetterPaper);
}

/*************************************************************************/

void setNormalOriginAndScale
(
  HP_StreamHandleType pStream
)
{
  HP_SetPageDefaultCTM_1(pStream);
  HP_SetPageOrigin_1(pStream, 100, 100);
  HP_SetPageScale_1(pStream, 1.0, 1.0);
}


/*************************************************************************/

void setFourUpOriginAndScale
(
  HP_StreamHandleType pStream,
  HP_UByte            block
)
{
  HP_SInt16 xOrigin;
  HP_SInt16 yOrigin;

  switch(block % 4)
  {
    case 0:
      xOrigin = 100;
      yOrigin = 100;
      break;
    case 1:
      xOrigin = 2400;
      yOrigin = 100;
      break;
    case 2:
      xOrigin = 100;
      yOrigin = 3150;
      break;
    case 3:
      xOrigin = 2400;
      yOrigin = 3150;
      break;
  }

  HP_SetPageDefaultCTM_1(pStream);
  HP_SetPageOrigin_1(pStream, xOrigin, yOrigin);
  HP_SetPageScale_1(pStream, 0.5, 0.5);
}

/*************************************************************************/

void endPage
(
  HP_StreamHandleType pStream
)
{
  HP_EndPage_2(pStream, 1);
}


/*************************************************************************/

void drawPages
(
  HP_StreamHandleType pStream,
  HP_pUByte           fontName,
  HP_UInt16           mapping,
  HP_UByte            firstBlock,
  HP_UByte            lastBlock,
  boolean             fourUp,
  boolean             vSub,
  boolean             vRot)
{
  HP_UInt16 block;
  HP_UByte  line;
  HP_UInt16 firstChar;
  HP_UInt16 i;
  HP_UInt16 characters[16];
  HP_SInt16 yOffset;
  HP_UInt16 escapements[16] = {217, 217, 217, 217, 217, 217, 217, 217,
                               217, 217, 217, 217, 217, 217, 217, 217};

  for(block=firstBlock;block<=lastBlock;block++)
  {
    if(!fourUp || (!(block % 4) || (block == firstBlock)))
    {
      beginPage(pStream);
    }
    if(fourUp)
    {
      setFourUpOriginAndScale(pStream, (HP_UByte)block);
    }
    else
    {
      setNormalOriginAndScale(pStream);
    }

    drawHeader(pStream, fontName);
    drawGrid(pStream);
    drawLabels(pStream, (HP_UByte)block);
    drawFooter(pStream, (HP_UByte)block, mapping);

    HP_SetFont_1(pStream,
                 (HP_pUByte)fontName,
                 (HP_UInt16)strlen((const char *)fontName),
                 150,
                 mapping);

    if(vSub)
    {
      HP_SetCharSubMode_1(pStream, (HP_pUByte)subArray, 1);
    }
    if(vRot)
    {
      HP_SetCharAttribute_1(pStream, HP_eWriteVertical);
    }

    for(line=0;line<16;line++)
    {
      yOffset = 1596 + (line * 236);
      firstChar = (256 * block) + (line * 16);
      for(i=0;i<16;i++)
      {
        characters[i] = firstChar + i;
      }
      HP_SetCursor_1(pStream, 935, yOffset);
      HP_Text_4(pStream, (HP_pUInt16)characters, 16, (HP_pSInt16)escapements,
        16);
      yOffset += 236;
    }

    if(!fourUp || (((block % 4) == 3) || (block == lastBlock)))
    {
      endPage(pStream);
    }
  }
}

/*************************************************************************/

void show_usage
(
  char *pgm
)
{
  fprintf(stderr,"%s V2.0\n", pgm);
  fprintf(stderr,
    "(c) Copyright Hewlett-Packard Company, 1998. All rights reserved.\n");
  fprintf(stderr,
    "USAGE: %s -o s -f s [-t s -a n -z n -c n -s n -4vrh]\n", pgm);
  fprintf(stderr,"Options:\n\
  -o s    Output File Name (required)\n\
  -f s    Device Font Name (OR Soft Font Name if used with -t) (required)\n\
  -m n    Mapping\n\
            Examples: 590 - Unicode      (18N)\n\
                      619 - Win31J       (19K)\n\
                      579 - GB-2312      (18C)\n\
                      643 - GBK          (20C)\n\
                      596 - BIG5         (18T)\n\
                      616 - KSC5601-1987 (19H)\n\
                      680 - UHC          (21H)\n\
            (default = 590)\n\
  -t n    TrueType font file to download and print with (optional)\n\
  -a n    First Block (default = 0)\n\
  -z n    Last Block  (default = 255)\n\
  -4      4-Up mode\n\
  -v      Perform vertical substitution\n\
  -r      Perform vertical rotation\n\
  -h      Print This Help Message\n\n");
}

/*************************************************************************/

void main(
  int   argc,
  char *argv[]
)
{
  HP_StreamHandleType     pStream;
  FILE                   *fontFile = NULL;
  char                   *fontFileName = NULL;
  FILE                   *outFile = NULL;
  char                   *outFileName = NULL;
  xlttlib_MemAccessType   memAccess;
  xlttlib_FontAccessType  fontAccess;
  xlttlib_FontInfoType    fontInfo;
  boolean                 ret;
  HP_UInt16               i=0;
  char                   *fontName = NULL;
  HP_UInt16               mapping = 590;
  boolean                 fourUp = FALSE;
  boolean                 vSub = FALSE;
  boolean                 vRot = FALSE;
  HP_UByte                firstBlock = 0x00;
  HP_UByte                lastBlock = 0xff;

  while(++i<argc)
  {
    if(argv[i][0] == '-')
    {
      switch(argv[i][1])
      {
        case 'o':              /* fileName */
          if(*(argv[i] + 2))
            outFileName = argv[i] + 2;
          else
            outFileName = argv[++i];
          break;
        case 'f':              /* fontName */
          if(*(argv[i] + 2))
            fontName = argv[i] + 2;
          else
            fontName = argv[++i];
          break;
        case 'm':              /* Symbol Set */
          if(*(argv[i]+2))
            mapping = atoi(argv[i]+2);
          else
            mapping = atoi(argv[++i]);
          break;
        case 't':              /* fontFileName */
          if(*(argv[i] + 2))
            fontFileName = argv[i] + 2;
          else
            fontFileName = argv[++i];
          break;
        case 'a':              /* Symbol Set */
          if(*(argv[i]+2))
            firstBlock = atoi(argv[i]+2);
          else
            firstBlock = atoi(argv[++i]);
          break;
        case 'z':              /* Symbol Set */
          if(*(argv[i]+2))
            lastBlock = atoi(argv[i]+2);
          else
            lastBlock = atoi(argv[++i]);
          break;
        case '4':              /* 4-Up */
          fourUp = TRUE;
          break;
        case 'v':              /* vertical substitution */
          vSub = TRUE;
          break;
        case 'r':              /* vertical rotation */
          vRot = TRUE;                
          break;
        case 'h':              /* help */
          show_usage(argv[0]);
          exit(0);
        default:
          fprintf(stderr,"%s: illegal option -%c\n",argv[0],argv[i][1]);
          show_usage(argv[0]);
          exit(0);
      }
    }
    else
    {
      show_usage(argv[0]);
      exit(1);
    }
  }

  if((outFileName == NULL) || (fontName == NULL))
  {
    show_usage(argv[0]);
    exit(0);
  }

  if(firstBlock > lastBlock)
  {
    fprintf(stderr, "\nError: First Block must be <= Last Block\n\n");
    exit(0);
  }
    
  if(fontFileName)
  {
    fontFile = fopen(fontFileName, "rb");
    if(fontFile == NULL)
    {
      fprintf(stderr, "client: Unable to open %s for reading\n",
        fontFileName);
      exit(1);
    }
  }

  outFile = fopen(outFileName, "wb");
  if(outFile == NULL)
  {
    fprintf(stderr, "client: Unable to create %s for writing\n", outFileName);
    exit(1);
  }

  memAccess.memAlloc = myAlloc;
  memAccess.memFree  = myFree;

  if(fontFile)
  {
    ret = xlttlib_Initialize(memAccess);
    if(!ret)
    {
      printf("client: xlttlib_Initialize failed\n");
      reportError(xlttlib_GetError());
      fclose(fontFile);
      fclose(outFile);
      exit(1);
    }
    fontAccess.seek          = mySeek;
    fontAccess.read          = myRead;
    fontAccess.clientDataPtr = (void *)fontFile;

    ret = xlttlib_RegisterFont(fontAccess, &fontInfo);
    if(!ret)
    {
      printf("client: xlttlib_RegisterFont failed\n");
      reportError(xlttlib_GetError());
      xlttlib_ShutDown();
      fclose(fontFile);
      fclose(outFile);
      exit(1);
    }
    printf("fontHandle = %u\n", fontInfo.fontHandle);
    printf("%s mapping = %u\n", fontFileName, fontInfo.mapping);
    if(fontInfo.mapping != mapping)
    {
      printf("client: %s can only be used with mapping = %d\n",
        fontFileName, fontInfo.mapping);
      xlttlib_ShutDown();
      fclose(fontFile);
      fclose(outFile);
      exit(1);
    }
  }

  pStream = HP_NewStream(OUT_BUFFER_MAX_SIZE, HP_FlushOutBuffer,
    (unsigned long)outFile);
  if(!pStream)
  {
    fprintf(stderr,"Error: Stream not initialized.\n");
    xlttlib_ShutDown();
    fclose(fontFile);
    fclose(outFile);
    exit(1);
  }

  HP_RawUByteArray(pStream, HP_UEL, 9);

  HP_RawUByteArray(pStream, HP_Enter_PCLXL, 26);

  HP_AddStreamHeader_1(pStream, 2, 0, HP_PCL_XL_Comment_Array, 43);

  HP_BeginSession_2(pStream, 600, 600, HP_eInch, HP_eBackChAndErrPage);

  HP_OpenDataSource_1(pStream, HP_eDefaultDataSource, HP_eBinaryLowByteFirst);

  if(fontFile)
  {
    ret = xlttlib_DownloadHeader(pStream, fontInfo.fontHandle,
      (HP_pUByte)fontName);
    if(!ret)
    {
      printf("client: xlttlib_DownloadHeader failed\n");
      reportError(xlttlib_GetError());
      xlttlib_ShutDown();
      fclose(fontFile);
      fclose(outFile);
      /* Flush UEL from the buffer out to the file */
      pStream->pFlushOutBuffer(pStream, pStream->cookie,
        pStream->HP_OutBuffer, pStream->HP_CurrentBufferLen);
      HP_FinishStream(pStream);
      exit(1);
    }

    ret = xlttlib_DownloadCharacters(pStream, fontInfo.fontHandle,
      (HP_pUByte)fontName, (firstBlock * 256), ((lastBlock * 256) + 255));
    if(!ret)
    {
      printf("client: xlttlib_DownloadCharacter failed\n");
      reportError(xlttlib_GetError());
      xlttlib_ShutDown();
      fclose(fontFile);
      fclose(outFile);
      /* Flush UEL from the buffer out to the file */
      pStream->pFlushOutBuffer(pStream, pStream->cookie,
        pStream->HP_OutBuffer, pStream->HP_CurrentBufferLen);
      HP_FinishStream(pStream);
      exit(1);
    }
  }

  drawPages(pStream, (HP_pUByte)fontName, mapping, firstBlock, lastBlock,
    fourUp, vSub, vRot);

  HP_CloseDataSource_1(pStream);

  HP_EndSession_1(pStream);

  HP_RawUByteArray(pStream, HP_UEL, 9);


  /* Flush UEL from the buffer out to the file */
  pStream->pFlushOutBuffer(pStream, pStream->cookie, pStream->HP_OutBuffer,
    pStream->HP_CurrentBufferLen);

  /* Free the memory allocated for the buffer */
  HP_FinishStream(pStream);

  fclose(outFile);

  if(fontFile)
  {
    xlttlib_ShutDown();
    fclose(fontFile);
  }

  printf("\nDONE\n\n");
  exit(0);
}

/*************************************************************************/

HP_StdFuncPrefix HP_FlushOutBuffer
(
  HP_StreamHandleType pStream,
  unsigned long       cookie,
  HP_pUByte           pOutBuffer,
  HP_SInt32           StreamBuffLen
)
{
  HP_SInt32  index;
  FILE      *outFile = (FILE *)cookie;

  index = 0;
  while(index != StreamBuffLen)
  {
    fputc(pOutBuffer[(HP_UInt16)index], outFile);
    index++;
  }

  pStream->HP_CurrentBufferLen = 0;  /* set outbuffer to empty */
}
