/*****************************************************************************

   (c) Copyright Hewlett-Packard Company, 1998.  All rights are
   reserved.  Copying or other reproduction of this program except
   for archival purposes is prohibited without the prior written
   consent of Hewlett-Packard Company.

                   RESTRICTED RIGHTS LEGEND

   Use, duplication, or disclosure by the Government is subject to
   restrictions as set forth in paragraph (b) (3) (B) of the Rights
   in Technical Data and Computer Software clause in DAR 7-104.9(a).

   HEWLETT-PACKARD COMPANY
   Boise, Idaho, USA

******************************************************************************/

/****************************** NOTICE ***************************************

HEWLETT-PACKARD COMPANY MAKES NO WARRANTIES, EXPRESS OR IMPLIED,  
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE
SOFTWARE OR TECHNICAL INFORMATION.   HEWLETT-PACKARD COMPANY DOES
NOT WARRANT, GUARANTEE OR MAKE ANY REPRESENTATIONS REGARDING THE 
USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR TECHNICAL 
INFORMATION IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY, 
CURRENTNESS, OR OTHERWISE.   THE ENTIRE RISK AS TO THE RESULTS AND 
PERFORMANCE OF THE SOFTWARE OR TECHNICAL INFORMATION IS ASSUMED
BY YOU.  The exclusion of implied warranties is not permitted by some 
jurisdictions.  The above exclusion may not apply to you.

IN NO EVENT WILL HEWLETT-PACKARD COMPANY BE LIABLE TO YOU FOR ANY 
CONSEQUENTIAL, INCIDENTAL OR INDIRECT DAMAGES (INCLUDING DAMAGES 
FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
INFORMATION AND THE LIKE)  ARISING OUT OF THE USE OR INABILITY TO USE 
THE SOFTWARE OR TECHNICAL INFORMATION EVEN IF HEWLETT-PACKARD HAS 
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  Because some jurisdictions 
do not allow the exclusion or limitation of liability for consequential
or incidental damages, the above limitations may not apply to you.  
Hewlett-Packard liability to you for actual damages from any cause whatsoever,
 and regardless of the form of the action (whether in contract, tort 
including negligence, product liability or otherwise), will be limited 
to US $50.

Copyright (c) 1998 Hewlett-Packard Company.  All rights reserved.

*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include "jetlib.h"
#include "debug.h"
#include "xlttlib.h"

/**************************************************************************/

#define TTF_VERSION 0x00010000L

#define CMAP_TAG    0x636d6170L
#define GLYF_TAG    0x676c7966L
#define HEAD_TAG    0x68656164L
#define HHEA_TAG    0x68686561L
#define HMTX_TAG    0x686d7478L
#define LOCA_TAG    0x6c6f6361L
#define MAXP_TAG    0x6d617870L
#define OS_2_TAG    0x4f532f32L

#define CVT__TAG    0x63767420L
#define FPGM_TAG    0x6670676dL
#define PREP_TAG    0x70726570L

#define MORT_TAG    0x6d6f7274L

#define VHEA_TAG    0x76686561L
#define VMTX_TAG    0x766d7478L

#define VERT_SUBTAG 0x76657274L

#define MS_PLATFORM_ID 3

#define SYMBOL_ENCODING        0
#define UNICODE_ENCODING       1
#define JAPANESE_ENCODING      2
#define SIMPCHINESE_ENCODING   3
#define TRADCHINESE_ENCODING   4
#define KOREANWANSUNG_ENCODING 5

#define ARG_1_AND_2_ARE_WORDS    0x01
#define WE_HAVE_A_SCALE          0x08
#define MORE_COMPONENTS          0x20
#define WE_HAVE_AN_X_AND_Y_SCALE 0x40
#define WE_HAVE_A_TWO_BY_TWO     0x80

#define FONT_HEADER_0_SIZE  8
#define DIR_HEADER_SIZE    12
#define SEG_HEADER_SIZE     6

#define PORTRAIT_ORIENTATION 0
#define TRUETYPE_FST         1

#define FORMAT1_CLASS1_HEADER_SIZE 10
#define FORMAT1_CLASS2_HEADER_SIZE 12

#define HIBYTE16(x) ((HP_UByte)(((HP_UInt16)x & 0xff00) >> 8))
#define LOBYTE16(x) ((HP_UByte)((HP_UInt16)x & 0x00ff))

#define HIWORD32(x) ((HP_UInt16)(((HP_UInt32)x & 0xffff0000L) >> 16))
#define LOWORD32(x) ((HP_UInt16)((HP_UInt32)x & 0x0000ffffL))

typedef struct TableDirTag
{
  HP_UInt32 cmapOffset;
  HP_UInt32 cmapLength;
  HP_UInt32 glyfOffset;
  HP_UInt32 glyfLength;
  HP_UInt32 headOffset;
  HP_UInt32 headLength;
  HP_UInt32 hheaOffset;
  HP_UInt32 hheaLength;
  HP_UInt32 hmtxOffset;
  HP_UInt32 hmtxLength;
  HP_UInt32 locaOffset;
  HP_UInt32 locaLength;
  HP_UInt32 maxpOffset;
  HP_UInt32 maxpLength;
  HP_UInt32 OS_2Offset;
  HP_UInt32 OS_2Length;
  HP_UInt32 cvt_Offset;
  HP_UInt32 cvt_Length;
  HP_UInt32 fpgmOffset;
  HP_UInt32 fpgmLength;
  HP_UInt32 prepOffset;
  HP_UInt32 prepLength;
  HP_UInt32 GSUBOffset;
  HP_UInt32 GSUBLength;
  HP_UInt32 mortOffset;
  HP_UInt32 mortLength;
  HP_UInt32 vheaOffset;
  HP_UInt32 vheaLength;
  HP_UInt32 vmtxOffset;
  HP_UInt32 vmtxLength;
}TableDirType;

typedef struct CachedCMAPSubtableTag
{
  HP_pUByte           subtablePtr; /* Points to allocated cmap subtable   */
                                   /* buffer. Filled with subtable data   */
      	                           /* following the version field. Note   */
                                   /* that the size field restricts the   */
                                   /* maximum size of a subtable to < 64k */ 
  HP_UInt16           length;
  HP_UInt16           format;
  xlttlib_MappingType mapping;
  HP_UInt16           numChars;
}CachedCMAPSubtableType;

typedef struct CachedMORTTableTag
{
  HP_UInt16 numEntries;
  HP_UInt16 tableSize;
  HP_pUByte tablePtr;   /* Points to allocated mort table buffer.        */
                        /* Filled with the LookupSingle array. Note that */
                        /* the length2 field restricts the maximum size  */
                        /* of the table to < 64k                         */
}CachedMORTTableType;

typedef struct FontTag *FontPtrType;

typedef struct FontTag
{
  FontPtrType            next;
  xlttlib_FontHandleType fontHandle;
  xlttlib_FontAccessType fontAccess;
  TableDirType           tableDir;
  HP_UInt16              maxGlyphID;
  HP_UInt16              unitsPerEm;
  HP_UInt16              indexToLocFormat;
  HP_SInt16              typoDescender;
  CachedCMAPSubtableType cachedCMAPSubtable;
  CachedMORTTableType    cachedMORTTable;
  HP_UInt16              numHorMetrics;
  HP_UInt16              numVerMetrics;
  HP_pUByte              glyphMarkingPtr;
  HP_UInt16              maxComponentDepth;
}FontType;

/*************************************************************************/
/* Glyph download types */

/*************************************************************************/
/* Global Variables */

static xlttlib_MemAccessType   gMemAccess;
static xlttlib_ErrorType       gError;
static boolean                 gInitialized = FALSE;
static xlttlib_FontHandleType  gFontHandleCounter = 0;
static FontType               *gFontDBHeadPtr;

#define DEFAULT_GC_SEG_SIZE 12
const HP_UByte DefaultGCSegment[DEFAULT_GC_SEG_SIZE] =
{
  'G','C',                /* Segment ID     */
  0x00, 0x00, 0x00, 0x06, /* Segment Size   */
  0x00, 0x00,             /* format         */
  0xff, 0xff,             /* default galley */
  0x00, 0x00              /* NumRanges      */
};

#define JAPANESE_GC_SEG_SIZE 18
const HP_UByte JapaneseGCSegment[JAPANESE_GC_SEG_SIZE] =
{
  'G','C',                /* Segment ID      */
  0x00, 0x00, 0x00, 0x0c, /* Segment Size    */
  0x00, 0x00,             /* format          */
  0xff, 0xff,             /* default galley  */
  0x00, 0x01,             /* NumRanges       */
  0x00, 0x00,             /* R1: upper-left  */
  0x00, 0xff,             /* R1: lower-right */
  0x00, 0xa5              /* R1: Galley      */
};

#define NON_UNICODE_VE_SEG_SIZE 12
const HP_UByte NonUnicodeVESegment[NON_UNICODE_VE_SEG_SIZE] =
{
  'V','E',                /* Segment ID   */
  0x00, 0x00, 0x00, 0x06, /* Segment Size */
  0x00,                   /* format       */
  0x01,                   /* NumRanges    */
  0x00, 0x00,             /* R1: First    */
  0x00, 0xff              /* R1: Last     */
};

/* The Unicode Consortium has not yet identified "vertical rotation" as a
   well-defined character property, so the following table was constructed
   using a "common sense" approach.
*/
#define UNICODE_VE_SEG_SIZE 28
const HP_UByte UnicodeVESegment[UNICODE_VE_SEG_SIZE] =
{
  'V','E',                /* Segment ID   */
  0x00, 0x00, 0x00, 0x16, /* Segment Size */
  0x00,                   /* format       */
  5,                      /* NumRanges    */
  0x00, 0x00,             /* R1: First    */
  0x2f, 0xff,             /* R1: Last     */
  0xfb, 0x00,             /* R2: First    */
  0xfe, 0x2f,             /* R2: Last     */
  0xfe, 0x50,             /* R3: First    */
  0xfe, 0xff,             /* R3: Last     */
  0xff, 0x60,             /* R4: First    */
  0xff, 0xdf,             /* R4: Last     */
  0xff, 0xe7,             /* R5: First    */
  0xff, 0xfe,             /* R5: Last     */
};

const HP_UByte NULLSegment[SEG_HEADER_SIZE] =
{
  0xFF, 0xFF, /* Segment ID   */
  0, 0, 0, 0  /* Segment Size */
};

#define VR_SEG_SIZE 10

/**************************************************************************/
/* Non-exported common function prototypes */

static void Error
(
  xlttlib_ErrorType error
);

static HP_pUByte MemAlloc
(
  HP_UInt16 numBytes
);

static void MemFree
(
  HP_pUByte ptr
);

static boolean Seek
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              offset
);

static boolean Read
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt16              numBytes,
  HP_pUByte              ptr
);

static boolean ReadHP_UInt16
(
  xlttlib_FontAccessType fontAccess,
  HP_pUInt16             ptr
);

static boolean ReadHP_UInt32
(
  xlttlib_FontAccessType fontAccess,
  HP_pUInt32             ptr
);

static boolean GetTableDir
(
  xlttlib_FontAccessType  fontAccess,
  TableDirType           *tableDirPtr
);

static boolean AddFontToDB
(
  FontType *fontPtr
);

static boolean DownloadGlyph
(
  HP_StreamHandleType  pStream,  
  FontType            *fontPtr,
  HP_UInt16            charCode,
  HP_UInt16            glyphID,
  HP_UInt16            componentDepth
);

static void DeleteFontDB
(
  void
);

static boolean DeleteFontFromDB
(
  xlttlib_FontHandleType fontHandle
);

static FontType *SearchFontDB
(
  xlttlib_FontHandleType fontHandle
);

static boolean CacheMORTTable
(
  xlttlib_FontAccessType  fontAccess,
  HP_UInt32               mortOffset,
  HP_UInt32               mortLength,
  CachedMORTTableType    *cachedMORTTablePtr
);

static boolean CacheCMAPSubtable
(
  xlttlib_FontAccessType  fontAccess,
  HP_UInt16               numGlyphs,
  HP_UInt32               cmapOffset,
  HP_UInt32               cmapLength,
  CachedCMAPSubtableType *cachedCMAPSubtablePtr
);

static boolean CalcNumChars4
(
  HP_UInt32  length,
  HP_pUByte  cmap4Ptr,
  HP_UInt16  maxGlyphID,
  HP_pUInt16 numCharsPtr
);

static HP_UInt16 Lookup4
(
  HP_UInt16 charCode,
  HP_pUByte cmap4Ptr,
  HP_UInt16 maxGlyphID
);

static boolean GetTypoDescender
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              OS_2Offset,
  HP_UInt32              OS_2Length,
  HP_pSInt16             typoDescenderPtr
);

static boolean GetNumMetrics
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              tableOffset,
  HP_UInt32              tableLength,
  HP_pUInt16             numMetricsPtr
);

static boolean GetHeadValues
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              headOffset,
  HP_UInt32              headLength,
  HP_pUInt16             unitePerEmPtr,
  HP_pUInt16             indexToLocFormatPtr
);

static boolean GetMAXPValues
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              maxpOffset,
  HP_UInt32              maxpLength,
  HP_pUInt16             numGlyphsPtr,
  HP_pUInt16             maxComponentDepthPtr
);

static boolean GetMetrics
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt16              glyphID,
  HP_UInt16              numMetrics,
  HP_UInt32              tableOffset,
  HP_pSInt16             bearingPtr,
  HP_pUInt16             advancePtr
);

static FontType *AllocateFont
(
  void
);

static void DeleteFont
(
  FontType *fontPtr
);

static boolean DownloadHeader
(
  HP_StreamHandleType  pStream,
  FontType            *font,
  HP_pUByte            fontName
);

static boolean DownloadGalleys
(
  HP_StreamHandleType  pStream,
  FontType            *fontPtr,
  HP_pUByte            fontName
);

void SendTTTable(
  HP_StreamHandleType    pStream,
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              tableOffset,
  HP_UInt32              tableLength  
);

void SendBuffer(
  HP_StreamHandleType pStream,
  HP_pUByte           bufferPtr,
  HP_UInt16           numBytes);

static boolean DownloadAllChars4
(
  HP_StreamHandleType  pStream,
  FontType            *fontPtr,
  HP_pUByte            fontName
);

static HP_UInt16 mortLookup
(
  CachedMORTTableType cachedMORTTable,
  HP_UInt16           glyphID
);

static HP_UInt16 CalcSizeOfMarkingBuffer
(
  HP_UInt16 maxGlyphID
);

static void ClearAllMarks
(
  HP_pUByte bitField,
  HP_UInt16 maxGlyphID
);

static void MarkGlyph
(
  HP_pUByte bitField,
  HP_UInt16 glyphID
);

static boolean IsGlyphMarked
(
  HP_pUByte bitField,
  HP_UInt16 glyphID
);

/**************************************************************************/

extern boolean xlttlib_Initialize
(
  xlttlib_MemAccessType memAccess
)
{
  DBG1("xlttlib: Enter xlttlib_Initialize\n");

  gMemAccess = memAccess;

  if((gMemAccess.memAlloc == NULL) ||
     (gMemAccess.memFree  == NULL))
  {
    Error(xlttlib_EInvalidMemAccess);
    DBG1("xlttlib: xlttlib_Initialize: EInvalidMemAccess\n");
    DBG1("xlttlib: Exit xlttlib_Initialize\n");
    return(FALSE);
  }

  gError = xlttlib_EUnknown;

  gFontDBHeadPtr = NULL;

  gInitialized = TRUE;

  DBG1("xlttlib: Exit xlttlib_Initialize\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_RegisterFont
(
  xlttlib_FontAccessType  fontAccess,
  xlttlib_FontInfoType   *fontInfoPtr
)
{
  boolean    ret;
  HP_UInt32  fileVersion;
  FontType  *fontPtr;
  HP_UInt16  numGlyphs;

  DBG1("xlttlib: Enter xlttlib_RegisterFont\n");

  if(gInitialized != TRUE)
  {
    Error(xlttlib_EInvalidCallSequence);
    DBG1("xlttlib: xlttlib_RegisterFont: EInvalidCallSequence\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if(gFontHandleCounter == 255)
  {
    Error(xlttlib_EOutOfFontHandles);
    DBG1("xlttlib: xlttlib_RegisterFont: EOutOfFontHandles\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if((fontAccess.seek == NULL) ||
     (fontAccess.read == NULL))
  {
    Error(xlttlib_EInvalidFontAccess);
    DBG1("xlttlib: xlttlib_RegisterFont: EInvalidFontAccess\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if(fontInfoPtr == NULL)
  {
    Error(xlttlib_EInvalidFontInfoPointer);
    DBG1("xlttlib: xlttlib_RegisterFont: EInvalidFontInfoPointer\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, 0);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_RegisterFont: seek Failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  ret = ReadHP_UInt32(fontAccess, &fileVersion);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_RegisterFont: read Failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if(fileVersion != TTF_VERSION)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: xlttlib_RegisterFont: EInvalidFont\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  DBG1("xlttlib: font looks like a TTF\n");

  /***********************************************************/
  /* Allocate memory for the font record */
  fontPtr = AllocateFont();
  if(!fontPtr)
  {
    DBG1("xlttlib: xlttlib_RegisterFont: AllocateFont failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /***********************************************************/
  /* Set the fonthandle and access fields */
  fontPtr->fontHandle = ++gFontHandleCounter;
  fontPtr->fontAccess = fontAccess;

  /***********************************************************/
  /* Read in the table directory */
  ret = GetTableDir(fontAccess, &(fontPtr->tableDir));
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: GetTableDir failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /***********************************************************/
  /* Get the numGlyphs field of the maxp table */
  ret = GetMAXPValues(
    fontAccess,
    fontPtr->tableDir.maxpOffset,
    fontPtr->tableDir.maxpLength,
    &numGlyphs,
    &fontPtr->maxComponentDepth);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: GetMAXPValues failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }
  fontPtr->maxGlyphID = numGlyphs - 1;

  /********************************************************************/
  /* Get the unitsPerEm and indexToLocFormat fields of the head table */
  ret = GetHeadValues(
    fontAccess,
    fontPtr->tableDir.headOffset,
    fontPtr->tableDir.headLength,
    &fontPtr->unitsPerEm,
    &fontPtr->indexToLocFormat);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: GetIndexToLocFormat failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /**************************************************************************/
  /* Check the validity of the loca table size against numGlyphs and
     indexToLocFormat
  */
  if(fontPtr->indexToLocFormat == 0)
  {
    if(fontPtr->tableDir.locaLength < (2 * (numGlyphs + 1)))
    {
      DeleteFont(fontPtr);
      Error(xlttlib_EInvalidFont);
      DBG1("xlttlib: xlttlib_RegisterFont: bad 'loca' table size\n");
      DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
      return(FALSE);
    }
  }
  else /* if(fontPtr->indexToLocFormat == 1) */
  {
    if(fontPtr->tableDir.locaLength < (4 * (numGlyphs + 1)))
    {
      DeleteFont(fontPtr);
      Error(xlttlib_EInvalidFont);
      DBG1("xlttlib: xlttlib_RegisterFont: bad 'loca' table size\n");
      DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
      return(FALSE);
    }
  }

  /***********************************************************/
  /* Get the sTypoDescender field of the OS/2 table */
  ret = GetTypoDescender(
    fontAccess,
    fontPtr->tableDir.OS_2Offset,
    fontPtr->tableDir.OS_2Length,
    &fontPtr->typoDescender);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: GetTypoDescender failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /**************************************************************************/
  /* Cache the first MS cmap subtable field of the head table and check the 
     validity of it's contents
  */
  ret = CacheCMAPSubtable(fontPtr->fontAccess, fontPtr->maxGlyphID,
    fontPtr->tableDir.cmapOffset, fontPtr->tableDir.cmapLength,
    &fontPtr->cachedCMAPSubtable);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: CacheCMAPSubtable failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /**************************************************************************/
  /* Get NumHorMetrics from hhea table and check validity of the hmtx table */
  ret = GetNumMetrics(
    fontAccess,
    fontPtr->tableDir.hheaOffset,
    fontPtr->tableDir.hheaLength,
    &fontPtr->numHorMetrics);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: GetNumMetrics failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if(fontPtr->tableDir.hmtxLength < (4 * fontPtr->numHorMetrics) +
     (2 * (numGlyphs - fontPtr->numHorMetrics)))
  {
    Error(xlttlib_EInvalidFont);
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: invalid hmtxLength\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  if(fontPtr->tableDir.vheaLength)
  {
    /**********************************************************************/
    /* Get NumVerMetrics from hhea table and check validity of vmtx table */
    ret = GetNumMetrics(
      fontAccess,
      fontPtr->tableDir.vheaOffset,
      fontPtr->tableDir.vheaLength,
      &fontPtr->numVerMetrics);
    if(!ret)
    {
      DeleteFont(fontPtr);
      DBG1("xlttlib: xlttlib_RegisterFont: GetNumMetrics failed\n");
      DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
      return(FALSE);
    }

    if(fontPtr->tableDir.vmtxLength < (4 * fontPtr->numVerMetrics) +
       (2 * (numGlyphs - fontPtr->numVerMetrics)))
    {
      Error(xlttlib_EInvalidFont);
      DeleteFont(fontPtr);
      DBG1("xlttlib: xlttlib_RegisterFont: invalid vmtxLength\n");
      DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
      return(FALSE);
    }
  }

  /***************************************************************/
  /* Cache the mort table if there is one and check its validity */
  if((fontPtr->tableDir.mortOffset != 0) &&
     (fontPtr->tableDir.mortLength != 0))
  {
    ret = CacheMORTTable(
      fontAccess,
      fontPtr->tableDir.mortOffset,
      fontPtr->tableDir.mortLength,
      &(fontPtr->cachedMORTTable));
    if(!ret)
    {
      DeleteFont(fontPtr);
      DBG1("xlttlib:xlttlib_RegisterFont: CacheMORTTable failed\n");
      DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
      return(FALSE);
    }
  }

  /*********************************************************************/
  /* Allocate memory for a bit-field used to mark glyphs as downloaded */
  fontPtr->glyphMarkingPtr =
    MemAlloc(CalcSizeOfMarkingBuffer(fontPtr->maxGlyphID));
  if(!fontPtr->glyphMarkingPtr)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: AllocateFont failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }
  ClearAllMarks(fontPtr->glyphMarkingPtr, fontPtr->maxGlyphID);

  /*************************************************************/
  /* Add the font record to the database */
  ret = AddFontToDB(fontPtr);
  if(!ret)
  {
    DeleteFont(fontPtr);
    DBG1("xlttlib: xlttlib_RegisterFont: AddFontToDB failed\n");
    DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
    return(FALSE);
  }

  /*************************************************************/
  /* Fill in the fontInfo structure to pass back to the client */
  fontInfoPtr->fontHandle = fontPtr->fontHandle;
  fontInfoPtr->mapping = fontPtr->cachedCMAPSubtable.mapping;

  DBG1("xlttlib: Exit xlttlib_RegisterFont\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_DownloadEntireFont
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName
)
{
  FontType *fontPtr;
  boolean   ret;

  DBG1("xlttlib: Enter xlttlib_DownloadEntireFont\n");

  fontPtr = SearchFontDB(fontHandle);
  if(!fontPtr)
  {
    Error(xlttlib_EInvalidFontHandle);
    DBG1("xlttlib: xlttlib_DownloadEntireFont: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit xlttlib_DownloadEntireFont\n");
  }

  ret = DownloadHeader(pStream, fontPtr, fontName);
  if(ret != TRUE)
  {
    DBG1("xlttlib: xlttlib_DownloadEntireFont: DownloadHeader failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadEntireFont\n");
    return(FALSE);
  }

  ret = DownloadGalleys(pStream, fontPtr, fontName);
  if(ret != TRUE)
  {
    DBG1("xlttlib: xlttlib_DownloadEntireFont: DownloadGalleys failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadEntireFont\n");
    return(FALSE);
  }

  /* Download all characters */
  ret = DownloadAllChars4(pStream, fontPtr, fontName);
  if(ret != TRUE)
  {
    DBG1("xlttlib: xlttlib_DownloadEntireFont: DownloadAllChars4 failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadEntireFont\n");
    return(FALSE);
  }

  DBG1("xlttlib: Exit xlttlib_DownloadEntireFont\n");
  return(ret);
}

/**************************************************************************/

extern boolean xlttlib_DownloadHeader
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName
)
{
  FontType *fontPtr;
  boolean   ret;

  DBG1("xlttlib: Enter xlttlib_DownloadHeader\n");

  fontPtr = SearchFontDB(fontHandle);
  if(!fontPtr)
  {
    Error(xlttlib_EInvalidFontHandle);
    DBG1("xlttlib: xlttlib_DownloadHeader: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit xlttlib_DownloadHeader\n");
  }

  ret = DownloadHeader(pStream, fontPtr, fontName);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_DownloadHeader: DownloadHeader failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadHeader\n");
    return(FALSE);
  }

  ret = DownloadGalleys(pStream, fontPtr, fontName);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_DownloadHeader: DownloadGalleys failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadHeader\n");
    return(FALSE);
  }

  DBG1("xlttlib: Exit xlttlib_DownloadHeader\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_DownloadCharacters
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName,
  HP_UInt16              startCode,
  HP_UInt16              endCode
)
{
  boolean   ret;
  FontType *fontPtr;
  HP_UInt32 charCode;
  HP_UInt16 glyphID;
  boolean   beginSent;

  DBG1("xlttlib: Enter xlttlib_DownloadCharacters\n");
  DBG5("xlttlib: xlttlib_DownloadCharacters: fontHandle=%d fontName=%s startCode=%d endCode=%d\n", fontHandle, fontName, startCode, endCode);

  if(startCode > endCode)
  {
    Error(xlttlib_EInvalidCharRange);
    DBG1("xlttlib: xlttlib_DownloadCharacters: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit xlttlib_DownloadCharacters\n");
    return(FALSE);
  }

  fontPtr = SearchFontDB(fontHandle);
  if(!fontPtr)
  {
    Error(xlttlib_EInvalidFontHandle);
    DBG1("xlttlib: xlttlib_DownloadCharacters: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit xlttlib_DownloadCharacters\n");
    return(FALSE);
  }

  beginSent = FALSE;

  for(charCode=startCode;charCode<=endCode;charCode++)
  {
    glyphID = Lookup4((HP_UInt16)charCode,
      fontPtr->cachedCMAPSubtable.subtablePtr, fontPtr->maxGlyphID);

    if(glyphID != 0)
    {
      if(!beginSent)
      {
        HP_BeginChar_1(pStream, fontName,
          (HP_UInt16)strlen((const char *)fontName));
        beginSent = TRUE;
      }
  
      ret = DownloadGlyph(pStream, fontPtr, (HP_UInt16)charCode, glyphID, 0);
      if(!ret)
      {
        DBG1("xlttlib: xlttlib_DownloadCharacters: DownloadGlyph failed\n");
        DBG1("xlttlib: Exit xlttlib_DownloadCharacters\n");
        return(FALSE);
      }
    }
  }

  if(beginSent)
  {
    HP_EndChar_1(pStream);
  }

  DBG1("xlttlib: Exit xlttlib_DownloadCharacters\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_GetLinearScaledCharWidths
(
  xlttlib_FontHandleType fontHandle,
  HP_Real32              charSize, 
  boolean                verticalRotation,
  boolean                verticalSubstitution,
  HP_UInt16              startCode,
  HP_UInt16              endCode,
  HP_pSInt16             widthsPtr
)
{
  FontType  *fontPtr;
  HP_UInt32  charCode;
  HP_UInt16  glyphID;
  HP_SInt16  bearing;
  HP_UInt16  advance;
  boolean    useVerticalMetrics;
  boolean    ret;

  DBG1("xlttlib: Enter xlttlib_GetLinearScaledCharWidths\n");

  if(!widthsPtr)
  {
    Error(xlttlib_EInvalidWidthsPtr);
    DBG1("xlttlib: xlttlib_GetLinearScaledCharWidths: EInvalidWidthsPtr\n");
    DBG1("xlttlib: Exit xlttlib_GetLinearScaledCharWidths\n");
    return(FALSE);
  }

  if(startCode > endCode)
  {
    Error(xlttlib_EInvalidCharRange);
    DBG1("xlttlib: xlttlib_GetLinearScaledCharWidths: EInvalidCharRange\n");
    DBG1("xlttlib: Exit xlttlib_GetLinearScaledCharWidths\n");
    return(FALSE);
  }

  fontPtr = SearchFontDB(fontHandle);
  if(!fontPtr)
  {
    Error(xlttlib_EInvalidFontHandle);
    DBG1("xlttlib: xlttlib_GetLinearScaledCharWidths: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit xlttlib_GetLinearScaledCharWidths\n");
    return(FALSE);
  }

  /* Get the unscaled widths from the appropriate metrics table */
  for(charCode=startCode;charCode<=endCode;charCode++)
  {
    glyphID = Lookup4((HP_UInt16)charCode,
      fontPtr->cachedCMAPSubtable.subtablePtr, fontPtr->maxGlyphID);
    if(verticalSubstitution && fontPtr->cachedMORTTable.tablePtr)
    {
      glyphID = mortLookup(fontPtr->cachedMORTTable, glyphID);
    }

    useVerticalMetrics = FALSE;
    if(verticalRotation)
    {
      useVerticalMetrics = TRUE;
      if(fontPtr->cachedCMAPSubtable.mapping == xlttlib_EUnicodeMapping)
      {
        if((charCode <= 0x2fff) ||
           ((charCode >= 0xfb00) && (charCode <= 0xfe2f)) ||
           ((charCode >= 0xfe50) && (charCode <= 0xfeff)) ||
           ((charCode >= 0xff60) && (charCode <= 0xffdf)) ||
           ((charCode >= 0xffe7) && (charCode <= 0xfffe)))
        {
          useVerticalMetrics = FALSE;
        }
      }
      else
      {
        if(charCode <= 0x00ff)
	{
          useVerticalMetrics = FALSE;
        }
      }
    }

    if(useVerticalMetrics && fontPtr->tableDir.vmtxOffset)
    {
      ret = GetMetrics(fontPtr->fontAccess, glyphID, fontPtr->numVerMetrics,
        fontPtr->tableDir.vmtxOffset, &bearing, &advance);
    }
    else
    {
      ret = GetMetrics(fontPtr->fontAccess, glyphID, fontPtr->numHorMetrics,
        fontPtr->tableDir.hmtxOffset, &bearing, &advance);
    }
    if(!ret)
    {
      DBG1("xlttlib: xlttlib_GetLinearScaledCharWidths: EInvalidFont\n");
      DBG1("xlttlib: Exit xlttlib_GetLinearScaledCharWidths\n");
      return(FALSE);
    }

    widthsPtr[(HP_UInt16)charCode-startCode] =
      (HP_UInt16)((HP_UInt32)advance * (HP_UInt32)charSize /
      fontPtr->unitsPerEm);
  }

  DBG1("xlttlib: Exit xlttlib_GetLinearScaledCharWidths\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_UnregisterFont
(
  xlttlib_FontHandleType FontHandle
)
{
  boolean ret;

  DBG1("xlttlib: Enter xlttlib_UnregisterFont\n");

  if(gInitialized != TRUE)
  {
    Error(xlttlib_EInvalidCallSequence);
    DBG1("xlttlib: xlttlib_UnregisterFont: EInvalidCallSequence\n");
    DBG1("xlttlib: Exit xlttlib_UnregisterFont\n");
    return(FALSE);
  }

  ret = DeleteFontFromDB(FontHandle);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_UnregisterFont: DeleteFontFromDB failed\n");
    DBG1("xlttlib: Exit xlttlib_UnregisterFont\n");
    return(FALSE);
  }

  DBG1("xlttlib: Exit xlttlib_UnregisterFont\n");
  return(TRUE);
}

/**************************************************************************/

extern boolean xlttlib_ShutDown
(
)
{
  DBG1("xlttlib: Enter xlttlib_ShutDown\n");

  if(gInitialized != TRUE)
  {
    Error(xlttlib_EInvalidCallSequence);
    DBG1("xlttlib: xlttlib_ShutDown: EInvalidCallSequence\n");
    DBG1("xlttlib: Exit xlttlib_ShutDown\n");
    return(FALSE);
  }

  DeleteFontDB();
  gInitialized = FALSE;

  DBG1("xlttlib: Exit xlttlib_ShutDown\n");
  return(TRUE);
}

/**************************************************************************/

extern xlttlib_ErrorType xlttlib_GetError
(
  void
)
{
  DBG1("xlttlib: Enter xlttlib_GetError\n");

  if(gInitialized != TRUE)
  {
    Error(xlttlib_EInvalidCallSequence);
    DBG1("xlttlib: xlttlib_GetError: EInvalidCallSequence\n");
  }

  DBG1("xlttlib: Exit xlttlib_GetError\n");
  return(gError);
}

/**************************************************************************/

static void Error
(
  xlttlib_ErrorType error
)
{
  DBG1("xlttlib: Enter Error\n");
  
  gError = error;
#ifdef DEBUG
  switch(gError)
  {
    case xlttlib_EInvalidMemAccess:
      DBG1("xlttlib: Error: EInvalidMemAccess\n");
      break;
    case xlttlib_EInvalidCallSequence:
      DBG1("xlttlib: Error: EInvalidCallSequence\n");
      break;
    case xlttlib_EInvalidFontInfoPointer:
      DBG1("xlttlib: Error: EInvalidFontInfoPointer\n");
      break;
    case xlttlib_EInvalidFontAccess:
      DBG1("xlttlib: Error: EInvalidFontAccess\n");
      break;
    case xlttlib_EInvalidFont:
      DBG1("xlttlib: Error: EInvalidFont\n");
      break;
    case xlttlib_EOutOfFontHandles:
      DBG1("xlttlib: Error: EOutOfFontHandles\n");
      break;
    case xlttlib_EInvalidFontHandle:
      DBG1("xlttlib: Error: EInvalidFontHandle\n");
      break;
    case xlttlib_EInvalidCharRange:
      DBG1("xlttlib: Error: EInvalidCharRange\n");
      break;
    case xlttlib_EInvalidWidthsPtr:
      DBG1("xlttlib: Error: EInvalidWidthsPtr\n");
      break;
    case xlttlib_ESeekError:
      DBG1("xlttlib: Error: ESeekError\n");
      break;
    case xlttlib_EReadError:
      DBG1("xlttlib: Error: EReadError\n");
      break;
    case xlttlib_EMemAllocError:
      DBG1("xlttlib: Error: EMemAllocError\n");
      break;
    case xlttlib_EUnknown:
    default:
      DBG2("xlttlib: Unknown error (0x%X)\n", error);
  }
#endif

  DBG1("xlttlib: Exit Error\n");
}

/**************************************************************************/

static HP_pUByte MemAlloc
(
  HP_UInt16 numBytes
)
{
  HP_pUByte ptr;

  DBG1("xlttlib: Enter MemAlloc\n");

  ptr = gMemAccess.memAlloc(numBytes);
  if(!ptr)
  {
    Error(xlttlib_EMemAllocError);
    DBG1("xlttlib: MemAlloc: EMemAllocError\n");
    DBG1("xlttlib: Exit MemAlloc\n");
    return(NULL);
  }

  DBG2("xlttlib: Allocated buffer at 0x%X\n", ptr); 
  DBG1("xlttlib: Exit MemAlloc\n");
  return(ptr);
}

/**************************************************************************/

static void MemFree
(
  HP_pUByte ptr
)
{
  DBG1("xlttlib: Enter MemFree\n");

  DBG2("xlttlib: Freeing buffer at 0x%X\n", ptr); 
  gMemAccess.memFree(ptr);

  DBG1("xlttlib: Exit MemFree\n");
}

/**************************************************************************/

static boolean Seek
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              offset
)
{
  boolean ret;

  DBG1("xlttlib: Enter Seek\n");

  ret = fontAccess.seek(fontAccess.clientDataPtr, offset);
  if(!ret)
  {
    Error(xlttlib_ESeekError);
    DBG1("xlttlib: Seek: ESeekError\n");
    DBG1("xlttlib: Exit Seek\n");
    return(FALSE);
  }

  DBG1("xlttlib: Exit Seek\n");
  return(TRUE);
}

/**************************************************************************/

static boolean Read
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt16              numBytes,
  HP_pUByte              ptr
)
{
  boolean ret;

  DBG1("xlttlib: Enter Read\n");

  ret = fontAccess.read(fontAccess.clientDataPtr, numBytes, ptr);
  if(!ret)
  {
    Error(xlttlib_EReadError);
    DBG1("xlttlib: Read: EReadError\n");
    DBG1("xlttlib: Exit Read\n");
    return(FALSE);
  }

  DBG1("xlttlib: Exit Read\n");
  return(TRUE);
}

/**************************************************************************/

static boolean ReadHP_UInt16
(
  xlttlib_FontAccessType fontAccess,
  HP_pUInt16             ptr
)
{
  HP_UByte bytes[2];
  boolean  ret;

  DBG1("xlttlib: Enter ReadHP_UInt16\n");

  ret = Read(fontAccess, 2, (HP_pUByte)bytes);
  if(!ret)
  {
    DBG1("xlttlib: ReadHP_UInt16: Read failed\n");
    DBG1("xlttlib: Exit ReadHP_UInt16\n");
    return(FALSE);
  }

  *ptr = (HP_UInt16)((HP_UInt16)bytes[0] << 8) |
                     (HP_UInt16)bytes[1];

  DBG1("xlttlib: Exit ReadHP_UInt16\n");
  return(TRUE);
}

/**************************************************************************/

static boolean ReadHP_UInt32
(
  xlttlib_FontAccessType fontAccess,
  HP_pUInt32             ptr
)
{
  HP_UByte bytes[4];
  boolean  ret;

  DBG1("xlttlib: Enter ReadHP_UInt16\n");

  ret = Read(fontAccess, 4, (HP_pUByte)bytes);
  if(!ret)
  {
    DBG1("xlttlib: ReadHP_UInt16: read failed\n");
    DBG1("xlttlib: Exit ReadHP_UInt16\n");
    return(FALSE);
  }

  *ptr =  (HP_UInt32)((HP_UInt32)bytes[0] << 24) |
                     ((HP_UInt32)bytes[1] << 16) |
                     ((HP_UInt32)bytes[2] <<  8) |
                      (HP_UInt32)bytes[3];

  DBG1("xlttlib: Exit ReadHP_UInt16\n");
  return(TRUE);
}

/**************************************************************************/

static boolean GetTableDir
(
  xlttlib_FontAccessType  fontAccess,
  TableDirType           *tableDirPtr
)
{
  HP_UInt16  numTags;
  HP_UInt16  searchRange;
  HP_UInt16  entrySelector;
  HP_UInt16  rangeShift;
  HP_UInt32  tag;
  HP_UInt32  checkSum;
  HP_UInt32  offset;
  HP_UInt32  length;
  HP_UInt16  i;
  boolean    ret;

  DBG1("xlttlib: Enter GetTableDir\n");

  ret = Seek(fontAccess, 4);
  if(!ret)
  {
    DBG1("xlttlib: GetTableDir: Seek failed\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &numTags);
  if(!ret)
  {
    DBG1("xlttlib: GetTableDir: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &searchRange);
  if(!ret)
  {
    DBG1("xlttlib: GetTableDir: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &entrySelector);
  if(!ret)
  {
    DBG1("xlttlib: GetTableDir: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &rangeShift);
  if(!ret)
  {
    DBG1("xlttlib: GetTableDir: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  for(i=0;i<numTags;i++)
  {
    ret = ReadHP_UInt32(fontAccess, &tag);
    if(!ret)
    {
      DBG1("xlttlib: GetTableDir: ReadHP_UInt32 failed\n");
      DBG1("xlttlib: Exit GetTableDir\n");
      return(FALSE);
    }

    ret = ReadHP_UInt32(fontAccess, &checkSum);
    if(!ret)
    {
      DBG1("xlttlib: GetTableDir: ReadHP_UInt32 failed\n");
      DBG1("xlttlib: Exit GetTableDir\n");
      return(FALSE);
    }

    ret = ReadHP_UInt32(fontAccess, &offset);
    if(!ret)
    {
      DBG1("xlttlib: GetTableDir: ReadHP_UInt32 failed\n");
      DBG1("xlttlib: Exit GetTableDir\n");
      return(FALSE);
    }

    ret = ReadHP_UInt32(fontAccess, &length);
    if(!ret)
    {
      DBG1("xlttlib: GetTableDir: ReadHP_UInt32 failed\n");
      DBG1("xlttlib: Exit GetTableDir\n");
      return(FALSE);
    }

    switch(tag)
    {
    case CMAP_TAG:
      tableDirPtr->cmapOffset = offset;
      tableDirPtr->cmapLength = length;
      DBG3("xlttlib: 'cmap' offset=%u length=%u\n", offset, length);
      break;
    case GLYF_TAG:
      tableDirPtr->glyfOffset = offset;
      tableDirPtr->glyfLength = length;
      DBG3("xlttlib: 'glyf' offset=%u length=%u\n", offset, length);
      break;
    case HEAD_TAG:
      tableDirPtr->headOffset = offset;
      tableDirPtr->headLength = length;
      DBG3("xlttlib: 'head' offset=%u length=%u\n", offset, length);
      break;
    case HHEA_TAG:
      tableDirPtr->hheaOffset = offset;
      tableDirPtr->hheaLength = length;
      DBG3("xlttlib: 'hhea' offset=%u length=%u\n", offset, length);
      break;
    case HMTX_TAG:
      tableDirPtr->hmtxOffset = offset;
      tableDirPtr->hmtxLength = length;
      DBG3("xlttlib: 'hmtx' offset=%u length=%u\n", offset, length);
      break;
    case LOCA_TAG:
      tableDirPtr->locaOffset = offset;
      tableDirPtr->locaLength = length;
      DBG3("xlttlib: 'loca' offset=%u length=%u\n", offset, length);
      break;
    case MAXP_TAG:
      tableDirPtr->maxpOffset = offset;
      tableDirPtr->maxpLength = length;
      DBG3("xlttlib: 'maxp' offset=%u length=%u\n", offset, length);
      break;
    case OS_2_TAG:
      tableDirPtr->OS_2Offset = offset;
      tableDirPtr->OS_2Length = length;
      DBG3("xlttlib: 'OS/2' offset=%u length=%u\n", offset, length);
      break;
    case CVT__TAG:
      tableDirPtr->cvt_Offset = offset;
      tableDirPtr->cvt_Length = length;
      DBG3("xlttlib: 'cvt ' offset=%u length=%u\n", offset, length);
      break;
    case FPGM_TAG:
      tableDirPtr->fpgmOffset = offset;
      tableDirPtr->fpgmLength = length;
      DBG3("xlttlib: 'fpgm' offset=%u length=%u\n", offset, length);
      break;
    case PREP_TAG:
      tableDirPtr->prepOffset = offset;
      tableDirPtr->prepLength = length;
      DBG3("xlttlib: 'prep' offset=%u length=%u\n", offset, length);
      break;
    case MORT_TAG:
      tableDirPtr->mortOffset = offset;
      tableDirPtr->mortLength = length;
      DBG3("xlttlib: 'mort' offset=%u length=%u\n", offset, length);
      break;
    case VHEA_TAG:
      tableDirPtr->vheaOffset = offset;
      tableDirPtr->vheaLength = length;
      DBG3("xlttlib: 'vhea' offset=%u length=%u\n", offset, length);
      break;
    case VMTX_TAG:
      tableDirPtr->vmtxOffset = offset;
      tableDirPtr->vmtxLength = length;
      DBG3("xlttlib: 'vmtx' offset=%u length=%u\n", offset, length);
      break;
    default:
      DBG7("xlttlib: unknown tag '%c%c%c%c' offset=%u length=%u\n",
         ((tag <<  0) >> 24),
         ((tag <<  8) >> 24),
         ((tag << 16) >> 24),
         ((tag << 24) >> 24),
         offset,
         length);
      break;
    }
  }

  /* Make sure the required tables exist */
  if((tableDirPtr->cmapOffset == 0) || (tableDirPtr->cmapLength == 0) ||
     (tableDirPtr->glyfOffset == 0) || (tableDirPtr->glyfLength == 0) ||
     (tableDirPtr->headOffset == 0) || (tableDirPtr->headLength == 0) ||
     (tableDirPtr->hheaOffset == 0) || (tableDirPtr->hheaLength == 0) ||
     (tableDirPtr->hmtxOffset == 0) || (tableDirPtr->hmtxLength == 0) ||
     (tableDirPtr->locaOffset == 0) || (tableDirPtr->locaLength == 0) ||
     (tableDirPtr->maxpOffset == 0) || (tableDirPtr->maxpLength == 0) ||
     (tableDirPtr->OS_2Offset == 0) || (tableDirPtr->OS_2Length == 0))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: GetTableDir: EInvalidFont\n");
    DBG1("xlttlib: Exit GetTableDir\n");
    return(FALSE);
  }

  if(((tableDirPtr->cvt_Offset == 0) && (tableDirPtr->cvt_Length != 0)) ||
     ((tableDirPtr->cvt_Offset != 0) && (tableDirPtr->cvt_Length == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad cvt size/length\n");
    tableDirPtr->cvt_Offset = 0;
    tableDirPtr->cvt_Length = 0;
  }

  if(((tableDirPtr->fpgmOffset == 0) && (tableDirPtr->fpgmLength != 0)) ||
     ((tableDirPtr->fpgmOffset != 0) && (tableDirPtr->fpgmLength == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad fpgm size/length\n");
    tableDirPtr->fpgmOffset = 0;
    tableDirPtr->fpgmLength = 0;
  }

  if(((tableDirPtr->prepOffset == 0) && (tableDirPtr->prepLength != 0)) ||
     ((tableDirPtr->prepOffset != 0) && (tableDirPtr->prepLength == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad prep size/length\n");
    tableDirPtr->prepOffset = 0;
    tableDirPtr->prepLength = 0;
  }

  if(((tableDirPtr->mortOffset == 0) && (tableDirPtr->mortLength != 0)) ||
     ((tableDirPtr->mortOffset != 0) && (tableDirPtr->mortLength == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad mort size/length\n");
    tableDirPtr->mortOffset = 0;
    tableDirPtr->mortLength = 0;
  }

  if(((tableDirPtr->vheaOffset == 0) && (tableDirPtr->vheaLength != 0)) ||
     ((tableDirPtr->vheaOffset != 0) && (tableDirPtr->vheaLength == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad vhea size/length\n");
    tableDirPtr->vheaOffset = 0;
    tableDirPtr->vheaLength = 0;
  }

  if(((tableDirPtr->vmtxOffset == 0) && (tableDirPtr->vmtxLength != 0)) ||
     ((tableDirPtr->vmtxOffset != 0) && (tableDirPtr->vmtxLength == 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad vmtx size/length\n");
    tableDirPtr->vmtxOffset = 0;
    tableDirPtr->vmtxLength = 0;
  }

  if(((tableDirPtr->vheaOffset != 0) && (tableDirPtr->vmtxOffset == 0)) ||
     ((tableDirPtr->vheaOffset == 0) && (tableDirPtr->vmtxOffset != 0)))
  {
    DBG1("xlttlib: GetTableDir: Warning: bad vhea/vmtx combination\n");
    tableDirPtr->vheaOffset = 0;
    tableDirPtr->vheaLength = 0;
    tableDirPtr->vmtxOffset = 0;
    tableDirPtr->vmtxLength = 0;
  }

  DBG1("xlttlib: Exit GetTableDir\n");
  return(TRUE);
}

/**************************************************************************/

static FontType *AllocateFont()
{
  FontType *fontPtr;

  DBG1("xlttlib: Enter AllocateFont\n");

  fontPtr = (FontType *)MemAlloc(sizeof(FontType));
  if(!fontPtr)
  {
    DBG1("xlttlib: AllocateFont: MemAlloc failed\n");
    return(NULL);
  }

  fontPtr->next = NULL;
  fontPtr->fontHandle = 0;
  fontPtr->maxGlyphID = 0;
  fontPtr->typoDescender = 0;

  fontPtr->fontAccess.seek          = NULL;
  fontPtr->fontAccess.read          = NULL;
  fontPtr->fontAccess.clientDataPtr = NULL;

  fontPtr->cachedCMAPSubtable.subtablePtr = NULL;
  fontPtr->cachedCMAPSubtable.length    = 0;
  fontPtr->cachedCMAPSubtable.format    = 0;
  fontPtr->cachedCMAPSubtable.mapping   = 0;

  fontPtr->tableDir.cmapOffset = 0;
  fontPtr->tableDir.cmapLength = 0;
  fontPtr->tableDir.glyfOffset = 0;
  fontPtr->tableDir.glyfLength = 0;
  fontPtr->tableDir.headOffset = 0;
  fontPtr->tableDir.headLength = 0;
  fontPtr->tableDir.hheaOffset = 0;
  fontPtr->tableDir.hheaLength = 0;
  fontPtr->tableDir.hmtxOffset = 0;
  fontPtr->tableDir.hmtxLength = 0;
  fontPtr->tableDir.locaOffset = 0;
  fontPtr->tableDir.locaLength = 0;
  fontPtr->tableDir.maxpOffset = 0;
  fontPtr->tableDir.maxpLength = 0;
  fontPtr->tableDir.OS_2Offset = 0;
  fontPtr->tableDir.OS_2Length = 0;
  fontPtr->tableDir.cvt_Offset = 0;
  fontPtr->tableDir.cvt_Length = 0;
  fontPtr->tableDir.fpgmOffset = 0;
  fontPtr->tableDir.fpgmLength = 0;
  fontPtr->tableDir.prepOffset = 0;
  fontPtr->tableDir.prepLength = 0;
  fontPtr->tableDir.mortOffset = 0;
  fontPtr->tableDir.mortLength = 0;
  fontPtr->tableDir.vheaOffset = 0;
  fontPtr->tableDir.vheaLength = 0;
  fontPtr->tableDir.vmtxOffset = 0;
  fontPtr->tableDir.vmtxLength = 0;

  fontPtr->cachedMORTTable.numEntries = 0;
  fontPtr->cachedMORTTable.tableSize  = 0;
  fontPtr->cachedMORTTable.tablePtr   = NULL;

  fontPtr->glyphMarkingPtr = NULL;

  fontPtr->numHorMetrics = 0;

  fontPtr->numVerMetrics = 0;

  fontPtr->maxComponentDepth = 0;

  DBG1("xlttlib: Exit AllocateFont\n");
  return(fontPtr);
}

/**************************************************************************/

static void DeleteFont
(
  FontType *fontPtr
)
{
  DBG1("xlttlib: Enter DeleteFont\n");
  if(fontPtr->cachedCMAPSubtable.subtablePtr)
  {
    MemFree(fontPtr->cachedCMAPSubtable.subtablePtr);
  }

  if(fontPtr->cachedMORTTable.tablePtr)
  {
    MemFree(fontPtr->cachedMORTTable.tablePtr);
  }

  if(fontPtr->glyphMarkingPtr)
  {
    MemFree(fontPtr->glyphMarkingPtr);
  }

  MemFree((HP_pUByte)fontPtr);
  DBG1("xlttlib: Exit DeleteFont\n");
}

/**************************************************************************/

static boolean AddFontToDB
(
  FontType *FontPtr
)
{
  DBG1("xlttlib: Enter AddFontToDB\n");

  FontPtr->next = gFontDBHeadPtr;
  gFontDBHeadPtr = FontPtr;

  DBG1("xlttlib: Exit AddFontToDB\n");
  return(TRUE);
}

/**************************************************************************/

static void DeleteFontDB
(
  void
)
{
  FontType *fontPtr;

  DBG1("xlttlib: Enter DeleteFontDB\n");

  fontPtr = gFontDBHeadPtr;

  while(fontPtr)
  {
    gFontDBHeadPtr = gFontDBHeadPtr->next;
    DeleteFont(fontPtr);
    fontPtr = gFontDBHeadPtr;
  }

  DBG1("xlttlib: Exit DeleteFontDB\n");
}

/**************************************************************************/

static boolean DeleteFontFromDB
(
  xlttlib_FontHandleType fontHandle
)
{
  FontType *fontPtr;
  FontType *tmpFontPtr;

  DBG2("xlttlib: Enter DeleteFontFromDB: fontHandle=%d\n", fontHandle);

  if(gFontDBHeadPtr == NULL)
  {
    Error(xlttlib_EInvalidFontHandle);
    DBG1("xlttlib: DeleteFontFromDB: EInvalidFontHandle\n");
    DBG1("xlttlib: Exit DeleteFontFromDB\n");
    return(FALSE);
  }

  fontPtr = gFontDBHeadPtr;

  /* Handle case where it's the head node */

  if(fontPtr->fontHandle == fontHandle)
  {
    gFontDBHeadPtr = gFontDBHeadPtr->next;
    DeleteFont(fontPtr);
    DBG1("xlttlib: Exit DeleteFontFromDB\n");
    return(TRUE);
  }

  /* Handle case where it's NOT the head node */

  while(fontPtr->next)
  {
    if(fontPtr->next->fontHandle == fontHandle)
    {
      tmpFontPtr = fontPtr->next;
      fontPtr->next = fontPtr->next->next;
      DeleteFont(tmpFontPtr);
      DBG1("xlttlib: Exit DeleteFontFromDB\n");
      return(TRUE);
    }

    fontPtr = fontPtr->next;
  }

  /* didn't find it */

  Error(xlttlib_EInvalidFontHandle);
  DBG1("xlttlib: DeleteFontFromDB: EInvalidFontHandle\n");
  DBG1("xlttlib: Exit DeleteFontFromDB\n");
  return(FALSE);
}

/**************************************************************************/

static FontType *SearchFontDB
(
  xlttlib_FontHandleType fontHandle
)
{
  FontType *fontPtr;

  DBG2("xlttlib: Enter SearchFontDB: fontHandle=%d\n", fontHandle);

  fontPtr = gFontDBHeadPtr;

  while(fontPtr)
  {
    if(fontPtr->fontHandle == fontHandle)
    {
      /* found it */
      DBG1("xlttlib: Exit SearchFontDB\n");
      return(fontPtr);
    }

    fontPtr = fontPtr->next;
  }

  /* didn't find it */
  Error(xlttlib_EInvalidFontHandle);
  DBG1("xlttlib: SearchFontDB: EInvalidFontHandle\n");
  DBG1("xlttlib: Exit SearchFontDB\n");
  return(NULL);
}

/**************************************************************************/

static boolean CacheMORTTable
(
  xlttlib_FontAccessType  fontAccess,
  HP_UInt32               mortOffset,
  HP_UInt32               mortLength,
  CachedMORTTableType    *cachedMORTTablePtr
)
{
  boolean   ret;
  HP_UByte  mortHeader[76];
  HP_UInt16 lookupLength;

  DBG3("xlttlib: Enter CacheMORTTable: mortOffset=%d, mortLength=%d\n",
    mortOffset, mortLength);

  ret = Seek(fontAccess, mortOffset);
  if(!ret)
  {
    DBG1("xlttlib: CacheMORTTable: Seek failed\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  if(mortLength <= 76)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheMORTTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  ret = Read(fontAccess, 76, mortHeader);
  if(!ret)
  {
    DBG1("xlttlib: CacheMORTTable: Read failed\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  /* We only understand a very specific version of the 'mort' table */
  if((mortHeader[0]  != 0x00) || /* 1st constant */
     (mortHeader[1]  != 0x01) ||
     (mortHeader[2]  != 0x00) ||
     (mortHeader[3]  != 0x00) ||
     (mortHeader[4]  != 0x00) ||
     (mortHeader[5]  != 0x00) ||
     (mortHeader[6]  != 0x00) ||
     (mortHeader[7]  != 0x01) ||
     (mortHeader[8]  != 0x00) ||
     (mortHeader[9]  != 0x00) ||
     (mortHeader[10] != 0x00) ||
     (mortHeader[11] != 0x01) ||

     (mortHeader[16] != 0x00) || /* 2nd constant */
     (mortHeader[17] != 0x03) ||
     (mortHeader[18] != 0x00) ||
     (mortHeader[19] != 0x01) ||
     (mortHeader[20] != 0x00) ||
     (mortHeader[21] != 0x03) ||
     (mortHeader[22] != 0x00) ||
     (mortHeader[23] != 0x00) ||
     (mortHeader[24] != 0x00) ||
     (mortHeader[25] != 0x00) ||
     (mortHeader[26] != 0x00) ||
     (mortHeader[27] != 0x01) ||
     (mortHeader[28] != 0xFF) ||
     (mortHeader[29] != 0xFF) ||
     (mortHeader[30] != 0xFF) ||
     (mortHeader[31] != 0xFF) ||

     (mortHeader[32] != 0x00) || /* 3rd constant */
     (mortHeader[33] != 0x03) ||
     (mortHeader[34] != 0x00) ||
     (mortHeader[35] != 0x01) ||
     (mortHeader[36] != 0x00) ||
     (mortHeader[37] != 0x00) ||
     (mortHeader[38] != 0x00) ||
     (mortHeader[39] != 0x00) ||
     (mortHeader[40] != 0xFF) ||
     (mortHeader[41] != 0xFF) ||
     (mortHeader[42] != 0xFF) ||
     (mortHeader[43] != 0xFE) ||
     (mortHeader[44] != 0x00) ||
     (mortHeader[45] != 0x08) ||
     (mortHeader[46] != 0x00) ||
     (mortHeader[47] != 0x00) ||

     (mortHeader[48] != 0x00) || /* 4th constant */
     (mortHeader[49] != 0x00) ||
     (mortHeader[50] != 0x00) ||
     (mortHeader[51] != 0x00) ||
     (mortHeader[52] != 0x00) ||
     (mortHeader[53] != 0x00) ||
     (mortHeader[54] != 0x00) ||
     (mortHeader[55] != 0x00) ||

     (mortHeader[58] != 0x80) || /* 5th constant */
     (mortHeader[59] != 0x04) ||
     (mortHeader[60] != 0x00) ||
     (mortHeader[61] != 0x00) ||
     (mortHeader[62] != 0x00) ||
     (mortHeader[63] != 0x01) ||
     (mortHeader[64] != 0x00) ||
     (mortHeader[65] != 0x06) ||

     (mortHeader[66] != 0x00) || /* entry Size must be 4 */
     (mortHeader[67] != 0x04))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheMORTTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  cachedMORTTablePtr->numEntries =
    (HP_UInt16)((HP_UInt16)mortHeader[68] << 8) + mortHeader[69];

  lookupLength = cachedMORTTablePtr->numEntries * 4;

  if((cachedMORTTablePtr->numEntries == 0) ||
     (mortLength < (76 + lookupLength)))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheMORTTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  /* We will also add a sentinel pair of 0xFFFFs at the end, so allocate an
     extra 4 bytes (The printer firmware requires the sentinel)
  */
  cachedMORTTablePtr->tableSize = lookupLength + 4;

  cachedMORTTablePtr->tablePtr = MemAlloc(cachedMORTTablePtr->tableSize);
  if(cachedMORTTablePtr->tablePtr == NULL)
  {
    DBG1("xlttlib: CacheMORTTable: MemAlloc failed\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  ret = Read(fontAccess, lookupLength,
    cachedMORTTablePtr->tablePtr);
  if(!ret)
  {
    DBG1("xlttlib: CacheMORTTable: Read failed\n");
    DBG1("xlttlib: Exit CacheMORTTable\n");
    return(FALSE);
  }

  /* Add a sentinel pair of 0xFFFFs at the end of the table */
  cachedMORTTablePtr->tablePtr[lookupLength]   = 0xFF;
  cachedMORTTablePtr->tablePtr[lookupLength+1] = 0xFF;
  cachedMORTTablePtr->tablePtr[lookupLength+2] = 0xFF;
  cachedMORTTablePtr->tablePtr[lookupLength+3] = 0xFF;

  DBG1("xlttlib: Exit CacheMORTTable\n");
  return(TRUE);
}

/**************************************************************************/

static boolean CacheCMAPSubtable
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt16              maxGlyphID,
  HP_UInt32              cmapOffset,
  HP_UInt32              cmapLength,
  CachedCMAPSubtableType *cachedCMAPSubtablePtr
)
{
  xlttlib_MappingType mapping;
  HP_UInt16           version;
  HP_UInt16           numTables;
  HP_UInt16           platformID;
  HP_UInt16           specificID;
  HP_UInt32           subtableOffset;
  HP_UInt16           format;
  HP_UInt16           length;
  HP_pUByte        cmapSubtablePtr;
  HP_UInt16           numChars;
  boolean             ret;
  HP_UInt16           i;

  DBG4("xlttlib: Enter CacheCMAPSubtable: cmapOffset=%d, cmapLength=%d, maxGlyphID=%d\n", cmapOffset, cmapLength, maxGlyphID);

  ret = Seek(fontAccess, cmapOffset);
  if(!ret)
  {
    DBG1("xlttlib: CacheCMAPSubtable: Seek failed\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  if(cmapLength < 4)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheCMAPTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &version);
  if(!ret)
  {
    DBG1("xlttlib: CacheCMAPSubtable: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  if(version != 0)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheCMAPTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &numTables);
  if(!ret)
  {
    DBG1("xlttlib: CacheCMAPSubtable: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  if(numTables == 0)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheCMAPTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  if(cmapLength < (4 + (8 * numTables)))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CacheCMAPTable: EInvalidFont\n");
    DBG1("xlttlib: Exit CacheCMAPSubtable\n");
    return(FALSE);
  }

  for(i=0;i<numTables;i++)
  {
    ret = ReadHP_UInt16(fontAccess, &platformID);
    if(!ret)
    {
      DBG1("xlttlib:CacheCMAPSubtable: ReadHP_UInt16 failed\n");
      DBG1("xlttlib: Exit CacheCMAPSubtable\n");
      return(FALSE);
    }

    ret = ReadHP_UInt16(fontAccess, &specificID);
    if(!ret)
    {
      DBG1("xlttlib:CacheCMAPSubtable: ReadHP_UInt16 failed\n");
      DBG1("xlttlib: Exit CacheCMAPSubtable\n");
      return(FALSE);
    }

    ret = ReadHP_UInt32(fontAccess, &subtableOffset);
    if(!ret)
    {
      DBG1("xlttlib:CacheCMAPSubtable: ReadHP_UInt32 failed\n");
      DBG1("xlttlib: Exit CacheCMAPSubtable\n");
      return(FALSE);
    }

    if(platformID == MS_PLATFORM_ID)
    {
      switch(specificID)
      {
        case SYMBOL_ENCODING:
          mapping = xlttlib_ESymbolMapping;
          break;
        case UNICODE_ENCODING:
          mapping = xlttlib_EUnicodeMapping;
          break;
        case JAPANESE_ENCODING:
          mapping = xlttlib_EJapaneseMapping;
          break;
        case SIMPCHINESE_ENCODING:
          mapping = xlttlib_ESimpChineseMapping;
          break;
        case TRADCHINESE_ENCODING:
          mapping = xlttlib_ETradChineseMapping;
          break;
        case KOREANWANSUNG_ENCODING:
          mapping = xlttlib_EKoreanWansungMapping;
          break;
        default:
          Error(xlttlib_EInvalidFont);
          DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
          DBG1("xlttlib: Exit CacheCMAPSubtable\n");
          return(FALSE);
      }

      if(cmapLength < (subtableOffset + 6))
      {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = Seek(fontAccess, cmapOffset + subtableOffset);
      if(!ret)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = ReadHP_UInt16(fontAccess, &format);
      if(!ret)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      if(format != 4)
      {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = ReadHP_UInt16(fontAccess, &length);
      if(!ret)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      if(length < 26) /* Check For Minimum Possible Length */
      {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = ReadHP_UInt16(fontAccess, &version);
      if(!ret)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      if(version != 0)
      {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      if(cmapLength < (subtableOffset + length))
      {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CacheCMAPSubtable: EInvalidFont\n");
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      length -= 6;
      cmapSubtablePtr = MemAlloc(length);
      if(cmapSubtablePtr == NULL)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = Read(fontAccess, length, cmapSubtablePtr);
      if(!ret)
      {
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      ret = CalcNumChars4(length, cmapSubtablePtr, maxGlyphID, &numChars);
      if((!ret) || (numChars == 0))
      {
        DBG1("xlttlib:CacheCMAPSubtable: CalcNumChars4 failed\n");
        MemFree(cmapSubtablePtr);
        DBG1("xlttlib: Exit CacheCMAPSubtable\n");
        return(FALSE);
      }

      cachedCMAPSubtablePtr->subtablePtr = cmapSubtablePtr;
      cachedCMAPSubtablePtr->length      = length;
      cachedCMAPSubtablePtr->format      = format;
      cachedCMAPSubtablePtr->mapping     = mapping;
      cachedCMAPSubtablePtr->numChars    = numChars;

      DBG1("xlttlib: Exit CacheCMAPSubtable\n");
      return(TRUE);
    }
  }

  DBG1("xlttlib:CacheCMAPSubtable: No valid subtable\n");
  DBG1("xlttlib: Exit CacheCMAPSubtable\n");
  return(FALSE);
}

/**************************************************************************/

static boolean CalcNumChars4
(
  HP_UInt32  length,
  HP_pUByte  cmap4Ptr,
  HP_UInt16  maxGlyphID,
  HP_pUInt16 numCharsPtr
)
{
  HP_UInt16 segCountX2;
  HP_UInt16 segCount;
  HP_pUByte endPtr;
  HP_pUByte startPtr;
  HP_pUByte idDeltaPtr;
  HP_SInt16 idDelta;
  HP_pUByte idRangeOffsetAddress;
  HP_pUByte idRangeOffsetPtr;
  HP_UInt16 idRangeOffset;
  HP_pUByte glyphIDPtr;
  HP_UInt16 glyphID;
  HP_UInt16 endCode;
  HP_UInt16 startCode;
  HP_UInt32 charCode;
  HP_UInt16 numChars;
  HP_UInt16 i;

  DBG3("xlttlib: Enter CalcNumChars4: length=%d, maxGlyphID=%d\n", length,
    maxGlyphID);

  numChars = 0;

  segCountX2 = ((HP_UInt16)cmap4Ptr[0] << 8) + cmap4Ptr[1];
  segCount = segCountX2 / 2;
  
  if(length < (10 + (8 * segCount))) /* Check for Minimum Possible length */
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: CalcNumChars4: EInvalidFont\n");
    DBG1("xlttlib: Exit CalcNumChars4\n");
    return(0);
  }

  endPtr = cmap4Ptr + 8;
  startPtr = cmap4Ptr + 10 + (2 * segCount);
  idDeltaPtr = startPtr + segCountX2;
  idRangeOffsetPtr = idDeltaPtr + segCountX2;
  glyphIDPtr = idRangeOffsetPtr + segCountX2;

  for(i=0;i<segCount;i++)
  {
    endCode = ((HP_UInt16)endPtr[i * 2] << 8) + endPtr[(i * 2) + 1];
    startCode = ((HP_UInt16)startPtr[i * 2] << 8) + startPtr[(i * 2) + 1];

    if(endCode < startCode)
    {
        Error(xlttlib_EInvalidFont);
        DBG1("xlttlib: CalcNumChars4: EInvalidFont\n");
        DBG1("xlttlib: Exit CalcNumChars4\n");
        return(0);
    }

    idDelta = ((HP_UInt16)idDeltaPtr[i * 2] << 8) + idDeltaPtr[(i * 2) + 1];
    idRangeOffsetAddress = idRangeOffsetPtr + (i * 2);
    idRangeOffset = ((HP_UInt16)idRangeOffsetAddress[0] << 8) +
      idRangeOffsetAddress[1];

    if(idRangeOffset != 0)
    {
      for(charCode=startCode;charCode<=endCode;charCode++)
      {
        glyphIDPtr = idRangeOffset + (((HP_UInt16)charCode - startCode) * 2) +
          idRangeOffsetAddress;
        glyphID = ((HP_UInt16)glyphIDPtr[0] << 8) + glyphIDPtr[1];
        if(glyphID != 0)
        {
          glyphID += idDelta;
        }
        if(glyphID > maxGlyphID)
        {
          DBG1("xlttlib: CalcNumChars4: Warning: glyphID > maxGlyphID\n");
          DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
            glyphID, maxGlyphID);
          glyphID = 0;
        }
      }
    }
    else
    {
      for(charCode=startCode;charCode<=endCode;charCode++)
      {
        glyphID = (HP_UInt16)charCode + idDelta;
        if(glyphID > maxGlyphID)
        {
          DBG1("xlttlib: CalcNumChars4: Warning: glyphID > maxGlyphID\n");
          DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
            glyphID, maxGlyphID);
          glyphID = 0;
        }
      }
    }

    numChars += ((endCode - startCode) + 1);
  }

  *numCharsPtr = numChars;
  DBG2("xlttlib: CalcNumChars4: numChars = %d\n", numChars);
  DBG1("xlttlib: Exit CalcNumChars4\n");
  return(TRUE);
}

/**************************************************************************/

static HP_UInt16 Lookup4
(
  HP_UInt16 charCode,
  HP_pUByte cmap4Ptr,
  HP_UInt16 maxGlyphID
)
{
  HP_UInt16 segCountX2;
  HP_UInt16 segCount;
  HP_pUByte endPtr;
  HP_pUByte startPtr;
  HP_pUByte idDeltaPtr;
  HP_SInt16 idDelta;
  HP_pUByte idRangeOffsetAddress;
  HP_pUByte idRangeOffsetPtr;
  HP_UInt16 idRangeOffset;
  HP_pUByte glyphIDPtr;
  HP_UInt16 glyphID;
  HP_UInt16 endCode;
  HP_UInt16 startCode;
  HP_UInt16 i;

  DBG3("xlttlib: Enter Lookup4: charCode=%d maxGlyphID=%d\n", charCode,
    maxGlyphID);

  glyphID = 0;

  segCountX2 = ((HP_UInt16)cmap4Ptr[0] << 8) + cmap4Ptr[1];
  segCount = segCountX2 / 2;
  
  endPtr = cmap4Ptr + 8;
  startPtr = cmap4Ptr + 10 + (2 * segCount);
  idDeltaPtr = startPtr + segCountX2;
  idRangeOffsetPtr = idDeltaPtr + segCountX2;
  glyphIDPtr = idRangeOffsetPtr + segCountX2;

  for(i=0;i<segCount;i++)
  {
    endCode = ((HP_UInt16)endPtr[i * 2] << 8) + endPtr[(i * 2) + 1];
    startCode = ((HP_UInt16)startPtr[i * 2] << 8) + startPtr[(i * 2) + 1];

    if(charCode <= endCode)
    {
      if(charCode >= startCode)
      {
        idDelta = ((HP_UInt16)idDeltaPtr[i * 2] << 8) +
          idDeltaPtr[(i * 2) + 1];
        idRangeOffsetAddress = idRangeOffsetPtr + (i * 2);
        idRangeOffset = ((HP_UInt16)idRangeOffsetAddress[0] << 8) +
          idRangeOffsetAddress[1];

        if(idRangeOffset != 0)
        {
          glyphIDPtr = idRangeOffset + ((charCode - startCode) * 2) +
            idRangeOffsetAddress;
          glyphID = ((HP_UInt16)glyphIDPtr[0] << 8) + glyphIDPtr[1];
          if(glyphID != 0)
          {
            glyphID += idDelta;
          }
          if(glyphID > maxGlyphID)
	  {
            DBG1("xlttlib: Lookup4: Warning: glyphID > maxGlyphID\n");
            DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
              glyphID, maxGlyphID);
            glyphID = 0;
          }
        }
        else
        {
          glyphID = charCode + idDelta;
          if(glyphID > maxGlyphID)
	  {
            DBG1("xlttlib: Lookup4: Warning: glyphID > maxGlyphID\n");
            DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
              glyphID, maxGlyphID);
            glyphID = 0;
          }
        }
      }
      else
      {
        break;
      }
    }
  }

  DBG2("xlttlib: Lookup4: glyphID = %d\n", glyphID);
  DBG1("xlttlib: Exit Lookup4\n");
  return(glyphID);
}

/**************************************************************************/

static boolean GetTypoDescender
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              OS_2Offset,
  HP_UInt32              OS_2Length,
  HP_pSInt16             typoDescenderPtr
)
{
  boolean   ret;
  HP_UInt16 version;
  HP_SInt16 typoDescender;

  DBG3("xlttlib: Enter GetTypoDescender: OS_2Offset=%d, OS_2Length=%d\n",
    OS_2Offset, OS_2Length);

  if(OS_2Length < 78)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, OS_2Offset);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &version);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }
  DBG2("xlttlib: OS/2.version = 0x%04X\n", version);
  if((version != 0x0000) && (version != 0x0001))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, OS_2Offset + 70);
  if(!ret)
  {
    DBG1("xlttlib: GetTypoDescender: Seek failed\n");
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, (HP_UInt16 *)&typoDescender);
  if(!ret)
  {
    DBG1("xlttlib: GetTypoDescender: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetTypoDescender\n");
    return(FALSE);
  }

  *typoDescenderPtr = typoDescender;
  DBG2("xlttlib: GetTypoDescender: typoDescender = %d\n", typoDescender);
  DBG1("xlttlib: Exit GetTypoDescender\n");
  return(TRUE);
}

/**************************************************************************/

static boolean GetNumMetrics
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              tableOffset,
  HP_UInt32              tableLength,
  HP_pUInt16             numMetricsPtr
)
{
  boolean   ret;
  HP_UInt32 version;
  HP_UInt16 numMetrics;

  DBG3("xlttlib: Enter GetNumMetrics: tableOffset=%d, tableLength=%d\n",
    tableOffset, tableLength);

  if(tableLength < 36)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, tableOffset);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }

  ret = ReadHP_UInt32(fontAccess, &version);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }
  DBG2("xlttlib: version = 0x%08X\n", version);
  if((version != 0x00010000L))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, tableOffset + 34);
  if(!ret)
  {
    DBG1("xlttlib: GetNumMetrics: Seek failed\n");
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, (HP_UInt16 *)&numMetrics);
  if(!ret)
  {
    DBG1("xlttlib: GetNumMetrics: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetNumMetrics\n");
    return(FALSE);
  }

  *numMetricsPtr = numMetrics;
  DBG2("xlttlib: GetNumMetrics: numMetrics = %d\n", numMetrics);
  DBG1("xlttlib: Exit GetNumMetrics\n");
  return(TRUE);
}

/**************************************************************************/

static boolean GetHeadValues
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              headOffset,
  HP_UInt32              headLength,
  HP_pUInt16             unitsPerEmPtr,
  HP_pUInt16             indexToLocFormatPtr
)
{
  boolean   ret;
  HP_UInt32 version;
  HP_UInt16 unitsPerEm;
  HP_UInt16 indexToLocFormat;

  DBG3("xlttlib: Enter GetHeadValues: headOffset=%d headLength=%d\n",
    headOffset, headLength);

  if(headLength < 54)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, headOffset);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt32(fontAccess, &version);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  DBG2("xlttlib: GetHeadValues: version=0x%08X\n", version);
  if(version != 0x00010000L)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: GetHeadValues: EInvalidFont\n");
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, headOffset + 18);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &unitsPerEm);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, headOffset + 50);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &indexToLocFormat);
  if(!ret)
  {
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  if((indexToLocFormat != 0) && (indexToLocFormat != 1))
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetHeadValues\n");
    return(FALSE);
  }

  *unitsPerEmPtr = unitsPerEm;
  *indexToLocFormatPtr = indexToLocFormat;
  DBG2("xlttlib: GetHeadValues: indexToLocFormat = %d\n",
    indexToLocFormat);
  DBG1("xlttlib: Exit GetHeadValues\n");
  return(TRUE);
}

/**************************************************************************/

static boolean GetMAXPValues
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              maxpOffset,
  HP_UInt32              maxpLength,
  HP_pUInt16             numGlyphsPtr,
  HP_pUInt16             maxComponentDepthPtr
)
{
  boolean   ret;
  HP_UInt32 version;
  HP_UInt16 numGlyphs;
  HP_UInt16 maxComponentDepth;

  DBG3("xlttlib: Enter GetMAXPValues:  maxpOffset=%d, maxpLength=%d\n",
    maxpOffset, maxpLength);

  if(maxpLength < 32)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, maxpOffset);
  if(!ret)
  {
    DBG1("xlttlib: GetMAXPValues: Seek failed\n");
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt32(fontAccess, &version);
  if(!ret)
  {
    DBG1("xlttlib: GetMAXPValues: ReadHP_UInt32 failed\n");
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  if(version != 0x00010000L)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &numGlyphs);
  if(!ret)
  {
    DBG1("xlttlib: GetMAXPValues: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  if(numGlyphs == 0)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, maxpOffset + 30);
  if(!ret)
  {
    DBG1("xlttlib: GetMAXPValues: Seek failed\n");
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, &maxComponentDepth);
  if(!ret)
  {
    DBG1("xlttlib: GetMAXPValues: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetMAXPValues\n");
    return(FALSE);
  }

  *numGlyphsPtr = numGlyphs;
  *maxComponentDepthPtr = maxComponentDepth;
  DBG3("xlttlib: GetMAXPValues: numGlyphs=%d maxComponentDepth=%d\n",
    numGlyphs, maxComponentDepth);
  DBG1("xlttlib: Exit GetMAXPValues\n");
  return(TRUE);
}

/**************************************************************************/

static boolean GetMetrics
(
  xlttlib_FontAccessType fontAccess,
  HP_UInt16              glyphID,
  HP_UInt16              numMetrics,
  HP_UInt32              tableOffset,
  HP_pSInt16             bearingPtr,
  HP_pUInt16             advancePtr
)
{
  HP_UInt32 advanceOffset;
  HP_UInt32 bearingOffset;
  boolean   ret;

  DBG3("xlttlib: Enter GetMetrics: glyphID=%d numMetrics=%d\n", glyphID,
    numMetrics);

  if(glyphID < numMetrics)
  {
    advanceOffset = tableOffset + (HP_UInt32)glyphID * 4;
    bearingOffset = advanceOffset + 2;
  }
  else
  {
    advanceOffset = tableOffset + (HP_UInt32)(numMetrics - 1) * 4;
    bearingOffset = tableOffset + ((HP_UInt32)numMetrics * 4) +
      ((HP_UInt32)(glyphID - numMetrics) * 2);
  }

  ret = Seek(fontAccess, advanceOffset);
  if(!ret)
  {
    DBG1("xlttlib: GetMetrics: Seek failed\n");
    DBG1("xlttlib: Exit GetMetrics\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, advancePtr);
  if(!ret)
  {
    DBG1("xlttlib: GetMetrics: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetMetrics\n");
    return(FALSE);
  }

  ret = Seek(fontAccess, bearingOffset);
  if(!ret)
  {
    DBG1("xlttlib: GetMetrics: Seek failed\n");
    DBG1("xlttlib: Exit GetMetrics\n");
    return(FALSE);
  }

  ret = ReadHP_UInt16(fontAccess, (HP_pUInt16)bearingPtr);
  if(!ret)
  {
    DBG1("xlttlib: GetMetrics: ReadHP_UInt16 failed\n");
    DBG1("xlttlib: Exit GetMetrics\n");
    return(FALSE);
  }

  DBG3("xlttlib: Exit GetMetrics: advance=%d bearing=%d\n", *advancePtr,
    *bearingPtr);
  return(TRUE);
}

/**************************************************************************/

static boolean DownloadHeader
(
  HP_StreamHandleType  pStream,
  FontType            *fontPtr,
  HP_pUByte            fontName
)
{
  HP_SInt16           typoDescender;
  xlttlib_MappingType mapping;
  HP_UInt16           numChars;
  HP_UInt32           headPaddedLength;
  HP_UInt32           maxpPaddedLength;
  HP_UInt32           cvtPaddedLength;
  HP_UInt32           fpgmPaddedLength;
  HP_UInt32           prepPaddedLength;
  HP_UByte            numTables;
  HP_UInt16           dirLength;
  HP_UInt32           GTSegmentLength;
  HP_UByte            fontHeader0[FONT_HEADER_0_SIZE];
  HP_UByte            dirHeader[DIR_HEADER_SIZE];
  HP_UByte            GTSegmentHeader[SEG_HEADER_SIZE];
  HP_UByte            VTSegmentHeader[SEG_HEADER_SIZE];
  HP_UByte            VRSegment[VR_SEG_SIZE];
  HP_UByte            dirTable[160]; /* Maximum Possible needed */
  HP_UInt16           powerOf2;
  HP_UByte            entrySelector;
  HP_UInt16           searchRange;
  HP_UInt16           rangeShift;
  HP_UInt32           endOfTable;
  HP_UByte            entryOffset;

  DBG2("xlttlib: Enter DownloadHeader: fontName=%s\n", fontName);

  typoDescender = fontPtr->typoDescender;
  numChars = fontPtr->cachedCMAPSubtable.numChars;
  mapping = fontPtr->cachedCMAPSubtable.mapping;

  /*
  Calculate Padded Lengths (All tables must be DWORD aligned) for the
  tables that will be included in the GT segment

  There are 3 required tables and 3 optional tables:
    required: head, maxp, glyf (empty table)
    optional: cvt,  fpgm, prep
  */

  headPaddedLength = fontPtr->tableDir.headLength;
  if(headPaddedLength % 4)
    headPaddedLength += 4 - (headPaddedLength % 4);

  maxpPaddedLength = fontPtr->tableDir.maxpLength;
  if(maxpPaddedLength % 4)
    maxpPaddedLength += 4 - (maxpPaddedLength % 4);

  cvtPaddedLength = fontPtr->tableDir.cvt_Length;
  if(cvtPaddedLength % 4)
    cvtPaddedLength += 4 - (cvtPaddedLength % 4);

  fpgmPaddedLength = fontPtr->tableDir.fpgmLength;
  if(fpgmPaddedLength % 4)
    fpgmPaddedLength += 4 - (fpgmPaddedLength % 4);

  prepPaddedLength = fontPtr->tableDir.prepLength;
  if(prepPaddedLength % 4)
    prepPaddedLength += 4 - (prepPaddedLength % 4);

  /* calculate number of tables to be included in the GT segment */
  numTables = (HP_UByte)(3 + (fontPtr->tableDir.cvt_Length ? 1 : 0) +
                             (fontPtr->tableDir.fpgmLength ? 1 : 0) +
                             (fontPtr->tableDir.prepLength ? 1 : 0));

  /* calculate the overall size of the table directory for the GT segment */
  dirLength = (HP_UInt16)((HP_UInt16)numTables * 16);

  /* calculate the overall size of the GT segment */
  GTSegmentLength = (HP_UInt32)DIR_HEADER_SIZE  +
                    (HP_UInt32)dirLength        +
                               cvtPaddedLength  +
                               fpgmPaddedLength +
                               headPaddedLength +
                               maxpPaddedLength +
                               prepPaddedLength;

  DBG2("xlttlib: DownloadHeader: GTSegmentLength=%d\n",
    GTSegmentLength);

  /*********************************************
  Begin the download stream process
  WARNING: From this point on, we cannot recover from a failed access to
           the TrueType font file!!!
  */

  HP_BeginFontHeader_1(pStream, fontName,
    (HP_UInt16)strlen((const char *)fontName), 0);

  /**************************************************/
  /* construct and send the XL font format 0 header */

  fontHeader0[0] = 0;                    /* format      */
  fontHeader0[1] = PORTRAIT_ORIENTATION; /* orientation */
  fontHeader0[2] = HIBYTE16(mapping);    /* mapping     */
  fontHeader0[3] = LOBYTE16(mapping);
  fontHeader0[4] = TRUETYPE_FST;         /* FST         */
  fontHeader0[5] = 0;                    /* variety     */
  fontHeader0[6] = HIBYTE16(numChars);   /* numChars    */
  fontHeader0[7] = LOBYTE16(numChars);

  SendBuffer(pStream, (HP_pUByte)fontHeader0, FONT_HEADER_0_SIZE);

  /*********************************************/
  /* construct and send the GT segment header */

  GTSegmentHeader[0] = 'G';                    /* Seg ID */
  GTSegmentHeader[1] = 'T';
  GTSegmentHeader[2] = HIBYTE16(HIWORD32(GTSegmentLength)); /* Length */
  GTSegmentHeader[3] = LOBYTE16(HIWORD32(GTSegmentLength));
  GTSegmentHeader[4] = HIBYTE16(LOWORD32(GTSegmentLength));
  GTSegmentHeader[5] = LOBYTE16(LOWORD32(GTSegmentLength));

  SendBuffer(pStream, (HP_pUByte)GTSegmentHeader, SEG_HEADER_SIZE);

  /*********************************************/
  /* construct and send the 'dir' table header */

  powerOf2 = 1;
  entrySelector = 0;
  while(powerOf2 * 2 <= numTables)
  {
    powerOf2 = (HP_UInt16)(powerOf2 * 2);
    entrySelector++;
  }
  searchRange = (HP_UInt16)(powerOf2 * 16);
  rangeShift = (numTables * 16) - searchRange;

  dirHeader[0]  = 0;                   /* SFNT version */
  dirHeader[1]  = 1;
  dirHeader[2]  = 0;
  dirHeader[3]  = 0;
  dirHeader[4]  = HIBYTE16(numTables); /* numTables */
  dirHeader[5]  = LOBYTE16(numTables);
  dirHeader[6]  = HIBYTE16(searchRange);
  dirHeader[7]  = LOBYTE16(searchRange);
  dirHeader[8]  = HIBYTE16(entrySelector);
  dirHeader[9]  = LOBYTE16(entrySelector);
  dirHeader[10] = HIBYTE16(rangeShift);
  dirHeader[11] = LOBYTE16(rangeShift);

  SendBuffer(pStream, (HP_pUByte)&dirHeader, DIR_HEADER_SIZE);

  /*********************************************
  Construct and send the table directory for the GT segment. The subtables
  must be listed in alphabetical order.
  */
  endOfTable = DIR_HEADER_SIZE + dirLength;
  entryOffset = 0;

  if(fontPtr->tableDir.cvt_Length != 0)
  {
    dirTable[entryOffset + 0]  = 'c'; /* tag */
    dirTable[entryOffset + 1]  = 'v';
    dirTable[entryOffset + 2]  = 't';
    dirTable[entryOffset + 3]  = ' ';
    dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
    dirTable[entryOffset + 5]  = 0;
    dirTable[entryOffset + 6]  = 0;
    dirTable[entryOffset + 7]  = 0;
    dirTable[entryOffset + 8]  = HIBYTE16(HIWORD32(endOfTable)); /* offset */
    dirTable[entryOffset + 9]  = LOBYTE16(HIWORD32(endOfTable));
    dirTable[entryOffset + 10] = HIBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 11] = LOBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 12] = HIBYTE16(HIWORD32(cvtPaddedLength));
    dirTable[entryOffset + 13] = LOBYTE16(HIWORD32(cvtPaddedLength));
    dirTable[entryOffset + 14] = HIBYTE16(LOWORD32(cvtPaddedLength));
    dirTable[entryOffset + 15] = LOBYTE16(LOWORD32(cvtPaddedLength));
    entryOffset += 16;
    endOfTable += cvtPaddedLength;
  }
  if(fontPtr->tableDir.fpgmLength != 0)
  {
    dirTable[entryOffset + 0]  = 'f'; /* tag */
    dirTable[entryOffset + 1]  = 'p';
    dirTable[entryOffset + 2]  = 'g';
    dirTable[entryOffset + 3]  = 'm';
    dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
    dirTable[entryOffset + 5]  = 0;
    dirTable[entryOffset + 6]  = 0;
    dirTable[entryOffset + 7]  = 0;
    dirTable[entryOffset + 8]  = HIBYTE16(HIWORD32(endOfTable)); /* offset */
    dirTable[entryOffset + 9]  = LOBYTE16(HIWORD32(endOfTable));
    dirTable[entryOffset + 10] = HIBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 11] = LOBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 12] = HIBYTE16(HIWORD32(fpgmPaddedLength));
    dirTable[entryOffset + 13] = LOBYTE16(HIWORD32(fpgmPaddedLength));
    dirTable[entryOffset + 14] = HIBYTE16(LOWORD32(fpgmPaddedLength));
    dirTable[entryOffset + 15] = LOBYTE16(LOWORD32(fpgmPaddedLength));
    entryOffset += 16;
    endOfTable += fpgmPaddedLength;
  }

  dirTable[entryOffset + 0]  = 'g'; /* tag */
  dirTable[entryOffset + 1]  = 'd';
  dirTable[entryOffset + 2]  = 'i';
  dirTable[entryOffset + 3]  = 'r';
  dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
  dirTable[entryOffset + 5]  = 0;
  dirTable[entryOffset + 6]  = 0;
  dirTable[entryOffset + 7]  = 0;
  dirTable[entryOffset + 8]  = 0; /* offset must be 0 */
  dirTable[entryOffset + 9]  = 0;
  dirTable[entryOffset + 10] = 0;
  dirTable[entryOffset + 11] = 0;
  dirTable[entryOffset + 12] = 0; /* Length must be 0 */
  dirTable[entryOffset + 13] = 0;
  dirTable[entryOffset + 14] = 0;
  dirTable[entryOffset + 15] = 0;
  entryOffset += 16;

  dirTable[entryOffset + 0]  = 'h'; /* tag */
  dirTable[entryOffset + 1]  = 'e';
  dirTable[entryOffset + 2]  = 'a';
  dirTable[entryOffset + 3]  = 'd';
  dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
  dirTable[entryOffset + 5]  = 0;
  dirTable[entryOffset + 6]  = 0;
  dirTable[entryOffset + 7]  = 0;
  dirTable[entryOffset + 8]  = HIBYTE16(HIWORD32(endOfTable)); /* offset */
  dirTable[entryOffset + 9]  = LOBYTE16(HIWORD32(endOfTable));
  dirTable[entryOffset + 10] = HIBYTE16(LOWORD32(endOfTable));
  dirTable[entryOffset + 11] = LOBYTE16(LOWORD32(endOfTable));
  dirTable[entryOffset + 12] = HIBYTE16(HIWORD32(headPaddedLength));
  dirTable[entryOffset + 13] = LOBYTE16(HIWORD32(headPaddedLength));
  dirTable[entryOffset + 14] = HIBYTE16(LOWORD32(headPaddedLength));
  dirTable[entryOffset + 15] = LOBYTE16(LOWORD32(headPaddedLength));
  entryOffset += 16;
  endOfTable += headPaddedLength;

  dirTable[entryOffset + 0]  = 'm'; /* tag */
  dirTable[entryOffset + 1]  = 'a';
  dirTable[entryOffset + 2]  = 'x';
  dirTable[entryOffset + 3]  = 'p';
  dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
  dirTable[entryOffset + 5]  = 0;
  dirTable[entryOffset + 6]  = 0;
  dirTable[entryOffset + 7]  = 0;
  dirTable[entryOffset + 8]  = HIBYTE16(HIWORD32(endOfTable)); /* offset */
  dirTable[entryOffset + 9]  = LOBYTE16(HIWORD32(endOfTable));
  dirTable[entryOffset + 10] = HIBYTE16(LOWORD32(endOfTable));
  dirTable[entryOffset + 11] = LOBYTE16(LOWORD32(endOfTable));
  dirTable[entryOffset + 12] = HIBYTE16(HIWORD32(maxpPaddedLength));
  dirTable[entryOffset + 13] = LOBYTE16(HIWORD32(maxpPaddedLength));
  dirTable[entryOffset + 14] = HIBYTE16(LOWORD32(maxpPaddedLength));
  dirTable[entryOffset + 15] = LOBYTE16(LOWORD32(maxpPaddedLength));
  entryOffset += 16;
  endOfTable += maxpPaddedLength;

  if(fontPtr->tableDir.prepLength != 0)
  {
    dirTable[entryOffset + 0]  = 'p'; /* tag */
    dirTable[entryOffset + 1]  = 'r';
    dirTable[entryOffset + 2]  = 'e';
    dirTable[entryOffset + 3]  = 'p';
    dirTable[entryOffset + 4]  = 0;   /* dummy check sum */
    dirTable[entryOffset + 5]  = 0;
    dirTable[entryOffset + 6]  = 0;
    dirTable[entryOffset + 7]  = 0;
    dirTable[entryOffset + 8]  = HIBYTE16(HIWORD32(endOfTable)); /* offset */
    dirTable[entryOffset + 9]  = LOBYTE16(HIWORD32(endOfTable));
    dirTable[entryOffset + 10] = HIBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 11] = LOBYTE16(LOWORD32(endOfTable));
    dirTable[entryOffset + 12] = HIBYTE16(HIWORD32(prepPaddedLength));
    dirTable[entryOffset + 13] = LOBYTE16(HIWORD32(prepPaddedLength));
    dirTable[entryOffset + 14] = HIBYTE16(LOWORD32(prepPaddedLength));
    dirTable[entryOffset + 15] = LOBYTE16(LOWORD32(prepPaddedLength));
    entryOffset += 16;
    endOfTable += prepPaddedLength;
  }

  SendBuffer(pStream, (HP_pUByte)dirTable, dirLength);

  /* Send individual tables. Tables must be in alphabetical order. */
  if(fontPtr->tableDir.cvt_Length != 0)
  {
    SendTTTable(pStream, fontPtr->fontAccess, fontPtr->tableDir.cvt_Offset,
      fontPtr->tableDir.cvt_Length);
  }

  if(fontPtr->tableDir.fpgmLength != 0)
  {
    SendTTTable(pStream, fontPtr->fontAccess, fontPtr->tableDir.fpgmOffset,
      fontPtr->tableDir.fpgmLength);
  }

  SendTTTable(pStream, fontPtr->fontAccess, fontPtr->tableDir.headOffset,
    fontPtr->tableDir.headLength);

  SendTTTable(pStream, fontPtr->fontAccess, fontPtr->tableDir.maxpOffset,
    fontPtr->tableDir.maxpLength);

  if(fontPtr->tableDir.prepLength != 0)
  {
    SendTTTable(pStream, fontPtr->fontAccess, fontPtr->tableDir.prepOffset,
      fontPtr->tableDir.prepLength);
  }

  /* Send a Galley Character Segment */
  if(fontPtr->cachedCMAPSubtable.mapping == xlttlib_EJapaneseMapping)
  {
    SendBuffer(pStream, (HP_pUByte)JapaneseGCSegment, JAPANESE_GC_SEG_SIZE);
  }
  else
  {
    SendBuffer(pStream, (HP_pUByte)DefaultGCSegment, DEFAULT_GC_SEG_SIZE);
  }

  /* Send a Vertical Exclude (VE) segment */
  if(fontPtr->cachedCMAPSubtable.mapping == xlttlib_EUnicodeMapping)
  {
    /* Send the Unicode VE segment */
    SendBuffer(pStream, (HP_pUByte)UnicodeVESegment, UNICODE_VE_SEG_SIZE);
  }
  else
  {
    /* Send the Non-Unicode VE segment */
    SendBuffer(pStream, (HP_pUByte)NonUnicodeVESegment,
      NON_UNICODE_VE_SEG_SIZE);
  }

  /* Construct and send a Vertical Rotation (VR) segment */
  VRSegment[0] = 'V';  /* Segment ID = 'VR' */
  VRSegment[1] = 'R';
  VRSegment[2] = 0x00; /* Segment Size = 4 */
  VRSegment[3] = 0x00;
  VRSegment[4] = 0x00;
  VRSegment[5] = 0x04;
  VRSegment[6] = 0x00; /* Format = 0 */
  VRSegment[7] = 0x00;
  VRSegment[8] = HIBYTE16(typoDescender);
  VRSegment[9] = LOBYTE16(typoDescender);

  SendBuffer(pStream, (HP_pUByte)VRSegment, VR_SEG_SIZE);

  /* Sent a VT segment if the font has a 'mort' table */
  if(fontPtr->cachedMORTTable.tablePtr)
  {
    VTSegmentHeader[0] = 'V';  /* Segment ID = 'VT' */
    VTSegmentHeader[1] = 'T';
    VTSegmentHeader[2] =
      HIBYTE16(HIWORD32(fontPtr->cachedMORTTable.tableSize));
    VTSegmentHeader[3] =
      LOBYTE16(HIWORD32(fontPtr->cachedMORTTable.tableSize));
    VTSegmentHeader[4] =
      HIBYTE16(LOWORD32(fontPtr->cachedMORTTable.tableSize));
    VTSegmentHeader[5] =
      LOBYTE16(LOWORD32(fontPtr->cachedMORTTable.tableSize));
    SendBuffer(pStream, (HP_pUByte)VTSegmentHeader, SEG_HEADER_SIZE);
    SendBuffer(pStream, fontPtr->cachedMORTTable.tablePtr,
      fontPtr->cachedMORTTable.tableSize);
  }

  /* Send the NULL segment */
  SendBuffer(pStream, (HP_pUByte)&NULLSegment, SEG_HEADER_SIZE);

  /* End the download process */
  HP_EndFontHeader_1(pStream);

  DBG1("xlttlib: Exit DownloadHeader\n");
  return(TRUE);
}

/**************************************************************************/
void SendTTTable(
  HP_StreamHandleType    pStream,
  xlttlib_FontAccessType fontAccess,
  HP_UInt32              tableOffset,
  HP_UInt32              tableLength  
)
{
  boolean   ret;
  HP_pUByte tempBuffer;
  HP_UInt32 numBytesLeft;
  HP_UByte  numPadBytes;
  HP_UInt32 padBuffer = 0L;
  
  DBG3("xlttlib: Enter SendTTTable: tableOffset=%d, tableLength=%d\n",
    tableOffset, tableLength);

  if(tableLength > 0xFFFF)
  {
    tempBuffer = MemAlloc(0xFFFF);
  }
  else
  {
    tempBuffer = MemAlloc((HP_UInt16)tableLength);
  }
  if(!tempBuffer)
  {
    DBG1("xlttlib: SendTTTable: MemAlloc Failed\n");
    DBG1("xlttlib: Exit SendTTTable\n");
  }

  ret = Seek(fontAccess, tableOffset);
  if(!ret)
  {
    DBG1("xlttlib: SendTTTable: Seek Failed\n");
    DBG1("xlttlib: Exit SendTTTable\n");
  } 

  numBytesLeft = tableLength;
  
  while(numBytesLeft > 0xFFFF)
  {
    ret = Read(fontAccess, 0xFFFF, tempBuffer);
    if(!ret)
    {
      DBG1("xlttlib: SendTTTable: Read Failed\n");
      DBG1("xlttlib: Exit SendTTTable\n");
    }
    SendBuffer(pStream, tempBuffer, 0xFFFF);
    numBytesLeft -= 0xFFFF;
  }

  if(numBytesLeft > 0L)
  {
    ret = Read(fontAccess, (HP_UInt16)numBytesLeft, tempBuffer);
    if(!ret)
    {
      DBG1("xlttlib: SendTTTable: Read Failed\n");
      DBG1("xlttlib: Exit SendTTTable\n");
    }
    SendBuffer(pStream, tempBuffer, (HP_UInt16)numBytesLeft);
  }

  if(tableLength % 4)
  {
    numPadBytes = 4 - (tableLength % 4);
    SendBuffer(pStream, (HP_pUByte)&padBuffer, numPadBytes);
  }

  MemFree(tempBuffer);
  DBG1("xlttlib: Exit SendTTTable\n");
}

/**************************************************************************/

void SendBuffer(
  HP_StreamHandleType pStream,
  HP_pUByte           bufferPtr,
  HP_UInt16           numBytes)
{
  DBG2("xlttlib: Enter SendBuffer: numBytes=%d\n", numBytes);

  HP_ReadFontHeader_1(pStream, numBytes);
  HP_EmbeddedDataPrefix(pStream, numBytes);
  HP_RawHugeUByteArray(pStream, bufferPtr, numBytes);

  DBG1("xlttlib: Exit SendBuffer\n");
}

/**************************************************************************/

static boolean DownloadGalleys
(
  HP_StreamHandleType  pStream,
  FontType            *fontPtr,
  HP_pUByte            fontName
)
{
  boolean   ret;
  HP_UInt16 glyphID;

  DBG1("xlttlib: Enter DownloadGalleys\n");

  HP_BeginChar_1(pStream, fontName,
    (HP_UInt16)strlen((const char *)fontName));

  ret = DownloadGlyph(pStream, fontPtr, (HP_UInt16)0xFFFF, 0, 0);
  if(!ret)
  {
    DBG1("xlttlib: xlttlib_DownloadHeader: DownloadGlyph failed\n");
    DBG1("xlttlib: Exit xlttlib_DownloadCharacters\n");
    return(FALSE);
  }

  if(fontPtr->cachedCMAPSubtable.mapping == xlttlib_EJapaneseMapping)
  {
    glyphID = Lookup4(0x00A5, fontPtr->cachedCMAPSubtable.subtablePtr,
      fontPtr->maxGlyphID);

    ret = DownloadGlyph(pStream, fontPtr, (HP_UInt16)0x00A5, glyphID, 0);
    if(!ret)
    {
      DBG1("xlttlib: xlttlib_DownloadGalleys: DownloadGlyph failed\n");
      DBG1("xlttlib: Exit xlttlib_DownloadGalleys\n");
      return(FALSE);
    }
  }

  HP_EndChar_1(pStream);

  DBG1("xlttlib: Exit DownloadGalleys\n");
  return(TRUE);
}

/**************************************************************************/

static boolean DownloadAllChars4
(
  HP_StreamHandleType  pStream,
  FontType            *fontPtr,
  HP_pUByte            fontName
)
{
  HP_pUByte cmap4Ptr;
  HP_UInt16 segCountX2;
  HP_UInt16 segCount;
  HP_pUByte endPtr;
  HP_pUByte startPtr;
  HP_pUByte idDeltaPtr;
  HP_SInt16 idDelta;
  HP_pUByte idRangeOffsetAddress;
  HP_pUByte idRangeOffsetPtr;
  HP_UInt16 idRangeOffset;
  HP_pUByte glyphIDPtr;
  HP_UInt16 glyphID;
  HP_UInt16 endCode;
  HP_UInt16 startCode;
  HP_UInt16 i;
  HP_UInt32 charCode;
  boolean   ret;

  DBG2("xlttlib: Enter DownloadAllChars4: fontName=%s\n", fontName);

  cmap4Ptr = fontPtr->cachedCMAPSubtable.subtablePtr;
  segCountX2 = ((HP_UInt16)cmap4Ptr[0] << 8) + cmap4Ptr[1];
  segCount = segCountX2 / 2;
  endPtr = cmap4Ptr + 8;
  startPtr = cmap4Ptr + 10 + segCountX2;
  idDeltaPtr = startPtr + segCountX2;
  idRangeOffsetPtr = idDeltaPtr + segCountX2;
  glyphIDPtr = idRangeOffsetPtr + segCountX2;

  HP_BeginChar_1(pStream, fontName,
    (HP_UInt16)strlen((const char *)fontName));

  for(i=0;i<segCount;i++)
  {
    endCode = ((HP_UInt16)endPtr[i * 2] << 8) + endPtr[(i * 2) + 1];
    startCode = ((HP_UInt16)startPtr[i * 2] << 8) + startPtr[(i * 2) + 1];
    idDelta = ((HP_UInt16)idDeltaPtr[i * 2] << 8) + idDeltaPtr[(i * 2) + 1];
    idRangeOffsetAddress = idRangeOffsetPtr + (i * 2);
    idRangeOffset = ((HP_UInt16)idRangeOffsetAddress[0] << 8) +
      idRangeOffsetAddress[1];

    DBG3("xlttlib: DownloadAllChars4: endCode=%d startCode=%d\n", endCode,
      startCode);

    if(idRangeOffset != 0)
    {
      for(charCode=startCode;charCode<=endCode;charCode++)
      {
        glyphIDPtr = idRangeOffset + (((HP_UInt16)charCode - startCode) * 2) +
          idRangeOffsetAddress;
        glyphID = ((HP_UInt16)glyphIDPtr[0] << 8) + glyphIDPtr[1];
        if(glyphID != 0)
        {
          glyphID += idDelta;
        }
        if(glyphID > fontPtr->maxGlyphID)
	{
          DBG1("xlttlib: DownloadAllChars4: Warn: glyphID > maxGlyphID\n");
          DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
            glyphID, fontPtr->maxGlyphID);
          glyphID = 0;
	}
        if(glyphID != 0)
	{
          ret = DownloadGlyph(pStream, fontPtr, (HP_UInt16)charCode, glyphID,
            0);
          if(!ret)
          {
            DBG1("xlttlib: DownloadAllChars4: DownloadGlyph failed\n");
            DBG1("xlttlib: Exit DownloadAllChars4\n");
            return(FALSE);
          }
        }
      }
    }
    else
    {
      for(charCode=startCode;charCode<=endCode;charCode++)
      {
        glyphID = (HP_UInt16)charCode + idDelta;
        if(glyphID > fontPtr->maxGlyphID)
        {
          DBG1("xlttlib: DownloadAllChars4: Warn: glyphID > maxGlyphID\n");
          DBG4("    charCode=%d glyphID=%d maxGlyphID=%d\n", charCode,
            glyphID, fontPtr->maxGlyphID);
          glyphID = 0;
	}
        if(glyphID != 0)
	{
          ret = DownloadGlyph(pStream, fontPtr, (HP_UInt16)charCode, glyphID,
            0);
          if(!ret)
          {
            DBG1("xlttlib: DownloadAllChars4: DownloadGlyph failed\n");
            DBG1("xlttlib: Exit DownloadAllChars4\n");
            return(FALSE);
          }
        }
      }
    }
  }

  HP_EndChar_1(pStream);

  DBG1("xlttlib: Exit DownloadAllChars4\n");
  return(TRUE);
}

/**************************************************************************/

static HP_UInt16 mortLookup
(
  CachedMORTTableType cachedMORTTable,
  HP_UInt16           glyphID
)
{
  HP_UInt16 i;
  HP_UInt16 horizontalGlyphID;
  HP_UInt16 verticalGlyphID;

  DBG2("xlttlib: Enter mortLookup: glyphID=%d\n", glyphID);

  for(i=0;i<cachedMORTTable.numEntries;i++)
  {
    horizontalGlyphID = ((HP_UInt16)cachedMORTTable.tablePtr[i*4] << 8) +
      cachedMORTTable.tablePtr[(i*4)+1];
    if(horizontalGlyphID == glyphID)
    {
      verticalGlyphID = ((HP_UInt16)cachedMORTTable.tablePtr[(i*4)+2] << 8) +
        cachedMORTTable.tablePtr[(i*4)+3];
      DBG3("xlttlib: Exit mortLookup: glyphID=%d, verticalGlyphID=%d\n",
        glyphID, verticalGlyphID);
      return(verticalGlyphID);
    }
  }

  DBG1("xlttlib: Exit mortLookup: verticalGlyphID=horizontalGlyphID\n");
  return(glyphID);
}

/**************************************************************************/

static boolean DownloadGlyph
(
  HP_StreamHandleType  pStream,  
  FontType            *fontPtr,
  HP_UInt16            charCode,
  HP_UInt16            glyphID,
  HP_UInt16            componentDepth
)
{
  boolean                ret;
  xlttlib_FontAccessType fontAccess;
  HP_UInt32              locaEntryOffset;
  HP_UInt16              shortOffset;
  HP_UInt32              thisGlyphOffset;
  HP_UInt32              nextGlyphOffset;
  HP_UInt32              glyphSize;
  HP_pUByte              glyphPtr;
  HP_UInt16              embeddedDataSize;
  HP_UInt16              charDataSize;
  HP_UByte               format1Class1Header[FORMAT1_CLASS1_HEADER_SIZE];
  HP_UByte               format1Class2Header[FORMAT1_CLASS2_HEADER_SIZE];
  HP_UInt16              componentGlyphID;
  HP_pUByte              tempGlyphPtr;
  HP_UInt16              flag;
  HP_SInt16              numContours;
  HP_SInt16              leftSideBearing;
  HP_UInt16              advanceWidth;
  HP_SInt16              topSideBearing;
  HP_UInt16              advanceHeight;
  HP_UInt16              verticalGlyphID;

  DBG1("xlttlib: Enter DownloadGlyph\n");
  DBG3("xlttlib: DownloadGlyph: charCode=%d glyphID=%d\n", charCode,
    glyphID);

  if(componentDepth > fontPtr->maxComponentDepth)
  {
    Error(xlttlib_EInvalidFont);
    DBG1("xlttlib: DownloadGlyph: componentDepth > maxComponentDepth\n");
    DBG1("xlttlib: Exit DownloadGlyph\n");
    return(FALSE);
  } 

  MarkGlyph(fontPtr->glyphMarkingPtr, glyphID);

  fontAccess = fontPtr->fontAccess;

  if(fontPtr->indexToLocFormat == 0)
  {
    locaEntryOffset = fontPtr->tableDir.locaOffset + (glyphID * 2);
    ret = Seek(fontAccess, locaEntryOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: Seek Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    } 
    ret = ReadHP_UInt16(fontAccess, (HP_pUInt16)&shortOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: ReadHP_UInt16 Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    }
    thisGlyphOffset = (HP_UInt32)shortOffset * 2;

    ret = ReadHP_UInt16(fontAccess, (HP_pUInt16)&shortOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: ReadHP_UInt16 Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    }
    nextGlyphOffset = (HP_UInt32)shortOffset * 2;
  }
  else /* if(fontPtr->indexToLocFormat == 1) */
  {
    locaEntryOffset = fontPtr->tableDir.locaOffset + (glyphID * 4);

    ret = Seek(fontAccess, locaEntryOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: Seek Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    } 
    ret = ReadHP_UInt32(fontAccess, (HP_pUInt32)&thisGlyphOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: ReadHP_UInt32 Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    } 
    ret = ReadHP_UInt32(fontAccess, (HP_pUInt32)&nextGlyphOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: ReadHP_UInt32 Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    } 
  }

  if(nextGlyphOffset > thisGlyphOffset)
  {
    glyphSize = nextGlyphOffset - thisGlyphOffset;

    if(glyphSize > (0xFFFF - FORMAT1_CLASS2_HEADER_SIZE))
    {
      Error(xlttlib_EInvalidFont);
      DBG1("xlttlib: DownloadGlyph: EInvalidFont\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    }
  }
  else
  {
    glyphSize = 0;
  }

  ret = GetMetrics(fontAccess, glyphID, fontPtr->numHorMetrics,
    fontPtr->tableDir.hmtxOffset, &leftSideBearing, &advanceWidth);
  if(!ret)
  {
    DBG1("xlttlib: DownloadGlyph: GetMetrics Failed\n");
    DBG1("xlttlib: Exit DownloadGlyph\n");
    return(FALSE);;
  }
  

  if(fontPtr->tableDir.vmtxOffset)
  {
    ret = GetMetrics(fontAccess, glyphID, fontPtr->numVerMetrics,
      fontPtr->tableDir.vmtxOffset, &topSideBearing, &advanceHeight);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: GetMetrics Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);;
    }

    /***********************************************************/
    /* Start the character download process */
    embeddedDataSize = FORMAT1_CLASS2_HEADER_SIZE + (HP_UInt16)glyphSize;
    HP_ReadChar_1(pStream, charCode, embeddedDataSize);
    HP_EmbeddedDataPrefix(pStream, embeddedDataSize);

    charDataSize = embeddedDataSize - 2;

    /***************************************************************/
    /* Construct and send the XL character format 1 class 2 header */
    format1Class2Header[0]  = 1; /* format */
    format1Class2Header[1]  = 2; /* class */
    format1Class2Header[2]  = HIBYTE16(charDataSize);
    format1Class2Header[3]  = LOBYTE16(charDataSize);
    format1Class2Header[4]  = HIBYTE16(leftSideBearing);
    format1Class2Header[5]  = LOBYTE16(leftSideBearing); 
    format1Class2Header[6]  = HIBYTE16(advanceWidth);
    format1Class2Header[7]  = LOBYTE16(advanceWidth); 
    format1Class2Header[8]  = HIBYTE16(topSideBearing);
    format1Class2Header[9]  = LOBYTE16(topSideBearing); 
    format1Class2Header[10] = HIBYTE16(glyphID);
    format1Class2Header[11] = LOBYTE16(glyphID); 

    HP_RawHugeUByteArray(pStream, (HP_pUByte)&format1Class2Header,
      FORMAT1_CLASS2_HEADER_SIZE);
  }
  else
  {
    /***********************************************************/
    /* Start the character download process */
    embeddedDataSize = FORMAT1_CLASS1_HEADER_SIZE + (HP_UInt16)glyphSize;
    HP_ReadChar_1(pStream, charCode, embeddedDataSize);
    HP_EmbeddedDataPrefix(pStream, embeddedDataSize);

    charDataSize = embeddedDataSize - 2;

    /***************************************************************/
    /* Construct and send the XL character format 1 class 1 header */
    format1Class1Header[0] = 1; /* format */
    format1Class1Header[1] = 1; /* class */
    format1Class1Header[2] = HIBYTE16(charDataSize);
    format1Class1Header[3] = LOBYTE16(charDataSize);
    format1Class1Header[4] = HIBYTE16(leftSideBearing);
    format1Class1Header[5] = LOBYTE16(leftSideBearing); 
    format1Class1Header[6] = HIBYTE16(advanceWidth);
    format1Class1Header[7] = LOBYTE16(advanceWidth); 
    format1Class1Header[8] = HIBYTE16(glyphID);
    format1Class1Header[9] = LOBYTE16(glyphID); 

    HP_RawHugeUByteArray(pStream, (HP_pUByte)&format1Class1Header,
      FORMAT1_CLASS1_HEADER_SIZE);
  }

  /***********************************************************/
  /* Send the glyph data */
  if(glyphSize != 0)
  {
    glyphPtr = MemAlloc((HP_UInt16)glyphSize);
    if(!glyphPtr)
    {
      DBG1("xlttlib: DownloadGlyph: MemAlloc failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    }

    ret = Seek(fontAccess, fontPtr->tableDir.glyfOffset + thisGlyphOffset);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: Seek Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);
    } 

    ret = Read(fontAccess, (HP_UInt16)glyphSize, glyphPtr);
    if(!ret)
    {
      DBG1("xlttlib: DownloadGlyph: Read Failed\n");
      DBG1("xlttlib: Exit DownloadGlyph\n");
      return(FALSE);;
    }

    HP_RawHugeUByteArray(pStream, (HP_pCharHuge)glyphPtr, glyphSize);

    tempGlyphPtr = glyphPtr;
    numContours = (HP_SInt16)((HP_UInt16)tempGlyphPtr[0] << 8) +
      tempGlyphPtr[1];

    if(numContours < 0)
    {
      tempGlyphPtr += 10;
      do
      {
        flag = (HP_UInt16)((HP_UInt16)tempGlyphPtr[0] << 8) +
          tempGlyphPtr[1];
        tempGlyphPtr += 2;

        /***********************************************************/
	/* Download the sub-glyph */
        componentGlyphID = (HP_UInt16)((HP_UInt16)tempGlyphPtr[0] << 8) +
          tempGlyphPtr[1];

        if(componentGlyphID > fontPtr->maxGlyphID)
	{
          Error(xlttlib_EInvalidFont);
          DBG1("xlttlib: DownloadGlyph: componentGlyphID > maxGlyphID\n");
          DBG3("    componentGlyphID=%d maxGlyphID=%d\n", componentGlyphID,
            fontPtr->maxGlyphID);
          DBG1("xlttlib: DownloadGlyph: EInvalidFont\n");
          DBG1("xlttlib: Exit DownloadGlyph\n");
          return(FALSE);
	}
        if(!IsGlyphMarked(fontPtr->glyphMarkingPtr, componentGlyphID))
	{
          ret = DownloadGlyph(pStream, fontPtr, 0xFFFF, componentGlyphID,
            componentDepth + 1);
          if(!ret)
          {
            DBG1("xlttlib: DownloadGlyph: DownloadGlyph failed\n");
            DBG1("xlttlib: Exit DownloadGlyph\n");
            return(FALSE);
          }
        }

        tempGlyphPtr += 2;
        if(flag & ARG_1_AND_2_ARE_WORDS)
        {
          tempGlyphPtr += 4;
        }
        else
        {
          tempGlyphPtr += 2;
        }

        if(flag & WE_HAVE_A_TWO_BY_TWO)
        {
          tempGlyphPtr += 8;
        }
        else if(flag  & WE_HAVE_AN_X_AND_Y_SCALE)
        {
          tempGlyphPtr += 4;
        }
        else if(flag  & WE_HAVE_A_SCALE)
        {
          tempGlyphPtr += 2;
        }
      } while (flag & MORE_COMPONENTS);
    }

    /* If this is a base character glyph (componentDepth == 0), check to see
       if it has a vertical substitute and download it.
    */
    if(fontPtr->cachedMORTTable.tablePtr && (componentDepth == 0))
    {
      verticalGlyphID = mortLookup(fontPtr->cachedMORTTable, glyphID);
      if(verticalGlyphID != glyphID)
      {
        if(!IsGlyphMarked(fontPtr->glyphMarkingPtr, verticalGlyphID))
	{
          ret = DownloadGlyph(pStream, fontPtr, 0xFFFF, verticalGlyphID, 0);
          if(!ret)
          {
            DBG1("xlttlib: DownloadGlyph: DownloadGlyph failed\n");
            DBG1("xlttlib: Exit DownloadGlyph\n");
            return(FALSE);
          }
        }
      }
    }
    MemFree(glyphPtr);
  }

  DBG1("xlttlib: Exit DownloadGlyph\n");
  return(TRUE);
}

/**************************************************************************/

static HP_UInt16 CalcSizeOfMarkingBuffer
(
  HP_UInt16  maxGlyphID
)
{
  HP_UInt16 size;

  DBG2("xlttlib: Enter CalcSizeOfMarkingBuffer: maxGlyphID=%d\n", maxGlyphID);

  size = (maxGlyphID + 8) / 8;

  DBG2("xlttlib: CalcSizeOfMarkingBuffer: size=%d\n", size);
  DBG1("xlttlib: Exit CalcSizeOfMarkingBuffer\n");
  return(size);
}

/**************************************************************************/

static void ClearAllMarks
(
  HP_pUByte bitField,
  HP_UInt16 maxGlyphID
)
{
  HP_UInt16 i;
  HP_UInt16 size;

  DBG2("xlttlib: Enter ClearAllMarks: maxGlyphID=%d\n", maxGlyphID);

  size = CalcSizeOfMarkingBuffer(maxGlyphID);
  for(i=0;i<size;i++)
  {
    bitField[i] = 0;
  }

  DBG1("xlttlib: Exit ClearAllMarks\n");
}

/**************************************************************************/

static void MarkGlyph
(
  HP_pUByte bitField,
  HP_UInt16 glyphID
)
{
  HP_UInt16 byteToSet;
  HP_UByte  bitToSet;
  HP_UByte  mask;

  DBG2("xlttlib: Enter MarkGlyph: glyphID=%d\n", glyphID);

  byteToSet = glyphID / 8;
  bitToSet = glyphID % 8;

  mask = (0x01 << bitToSet);

  bitField[byteToSet] |= mask;
  DBG1("xlttlib: Exit MarkGlyph\n");
}

/**************************************************************************/

static boolean IsGlyphMarked
(
  HP_pUByte bitField,
  HP_UInt16 glyphID
)
{
  HP_UInt16 byteToTest;
  HP_UByte  bitToTest;
  HP_UByte  mask;
  boolean   ret;

  DBG2("xlttlib: Enter IsGlyphMarked: glyphID=%d\n", glyphID);

  byteToTest = glyphID / 8;
  bitToTest = glyphID % 8;
  mask = (0x01 << bitToTest);
  ret = FALSE;

  if(bitField[byteToTest] & mask)
  {
    ret = TRUE;
  }

  DBG2("xlttlib: IsGlyphMarked ret=%d\n", ret);
  DBG1("xlttlib: Exit IsGlyphMarked\n");
  return(ret);
}
