/*****************************************************************************

   (c) Copyright Hewlett-Packard Company, 1998.  All rights are
   reserved.  Copying or other reproduction of this program except
   for archival purposes is prohibited without the prior written
   consent of Hewlett-Packard Company.

                   RESTRICTED RIGHTS LEGEND

   Use, duplication, or disclosure by the Government is subject to
   restrictions as set forth in paragraph (b) (3) (B) of the Rights
   in Technical Data and Computer Software clause in DAR 7-104.9(a).

   HEWLETT-PACKARD COMPANY
   Boise, Idaho, USA

******************************************************************************/

/****************************** NOTICE ***************************************

HEWLETT-PACKARD COMPANY MAKES NO WARRANTIES, EXPRESS OR IMPLIED,  
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE
SOFTWARE OR TECHNICAL INFORMATION.   HEWLETT-PACKARD COMPANY DOES
NOT WARRANT, GUARANTEE OR MAKE ANY REPRESENTATIONS REGARDING THE 
USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR TECHNICAL 
INFORMATION IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY, 
CURRENTNESS, OR OTHERWISE.   THE ENTIRE RISK AS TO THE RESULTS AND 
PERFORMANCE OF THE SOFTWARE OR TECHNICAL INFORMATION IS ASSUMED
BY YOU.  The exclusion of implied warranties is not permitted by some 
jurisdictions.  The above exclusion may not apply to you.

IN NO EVENT WILL HEWLETT-PACKARD COMPANY BE LIABLE TO YOU FOR ANY 
CONSEQUENTIAL, INCIDENTAL OR INDIRECT DAMAGES (INCLUDING DAMAGES 
FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
INFORMATION AND THE LIKE)  ARISING OUT OF THE USE OR INABILITY TO USE 
THE SOFTWARE OR TECHNICAL INFORMATION EVEN IF HEWLETT-PACKARD HAS 
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  Because some jurisdictions 
do not allow the exclusion or limitation of liability for consequential
or incidental damages, the above limitations may not apply to you.  
Hewlett-Packard liability to you for actual damages from any cause whatsoever,
 and regardless of the form of the action (whether in contract, tort 
including negligence, product liability or otherwise), will be limited 
to US $50.

Copyright (c) 1998 Hewlett-Packard Company.  All rights reserved.

*****************************************************************************/

#ifndef XLTTLIB_H
#define XLTTLIB_H

#include "jetlib.h"

typedef unsigned char boolean;
#define TRUE 1
#define FALSE 0

/**************************************************************************
    If any of the xlttlib routines return failure, the client should call
    xlttlib_GetError which will return one of the following codes.
*/

typedef enum xlttlib_ErrorTag
{
  xlttlib_EUnknown,
  xlttlib_EInvalidMemAccess,
  xlttlib_EInvalidCallSequence,
  xlttlib_EInvalidFontInfoPointer,
  xlttlib_EInvalidFontAccess,
  xlttlib_EInvalidFont,
  xlttlib_EOutOfFontHandles,
  xlttlib_EInvalidFontHandle,
  xlttlib_EInvalidCharRange,
  xlttlib_EInvalidWidthsPtr,
  xlttlib_ESeekError,
  xlttlib_EReadError,
  xlttlib_EMemAllocError
}xlttlib_ErrorType;

/**************************************************************************
    Before initializing xlttlib, the client program must fill in the
    following structure. xlttlib will use the fields of this structure
    to allocate memory for its own temporary use.
*/

typedef struct xlttlib_MemAccessTag
{
  HP_pUByte (*memAlloc)(HP_UInt16 byteCount);
  void  (*memFree)(HP_pUByte dataPtr);
}xlttlib_MemAccessType;

/**************************************************************************
    Before registering each font, the client program must fill in the
    following structure. xlttlib will use the fields of this structure
    to access the font data. This allows the client to store the font
    wherever it wants - memory, disk, network, etc. The clientDataPtr
    will be passed back to the client anytime xlttlib accesses the font
    and may be used by the client to point to any data it would like.
*/

typedef struct xlttlib_FontAccessTag
{
  boolean (*seek)(void *clientDataPtr, HP_UInt32 offsetFromTop);
  boolean (*read)(void *clientDataPtr, HP_UInt16 byteCount, HP_pUByte buffer);
  void    *clientDataPtr;
}xlttlib_FontAccessType;

/**************************************************************************
    When a font is registered with xlttlib, it will be assigned a handle
    of the following type. xlttlib can only manage 254 fonts at a time.
*/

typedef HP_UByte xlttlib_FontHandleType;

/**************************************************************************
    xlttlib will search the 'cmap' table of every registered font. The
    first Microsoft subtable it finds will determine the "mapping" for
    that font. All character data for that font must be in the specified
    mapping. xlttlib currently supports the following mappings.
*/

typedef enum xlttlib_MappingTag
{
  xlttlib_ESymbolMapping = 621,        /* Symbol (U+F000-U+F0FF)   */
  xlttlib_EUnicodeMapping = 590,       /* Unicode                  */
  xlttlib_EJapaneseMapping = 619,      /* Win31J-DBCS (Shift-JIS)  */
  xlttlib_ESimpChineseMapping = 643,   /* GB2312-1980 or GBK       */
  xlttlib_ETradChineseMapping = 596,   /* Big5                     */
  xlttlib_EKoreanWansungMapping = 616  /* Wansung (KSC5601-1987)   */
}xlttlib_MappingType;

/**************************************************************************
    When a font is registered with xlttlib, the client must pass a pointer
    to the following structure for xlttlib to fill in. This structure will
    contain information that the client will need to be able to use the
    font.
*/

typedef struct xlttlib_FontInfoTag
{
  /* When a font is registered, it will be assigned a handle by xlttlib. */
  /* Use the handle in all subsequent calls to xlttlib.                  */

  xlttlib_FontHandleType fontHandle;

  /* xlttlib will search the 'cmap' table of every registered font. The   */
  /* first Microsoft subtable it finds will determine the "mapping" for   */
  /* that font. Strings must be passed to the HP_Text_* functions in this */
  /* form.                                                                */

  xlttlib_MappingType mapping;

}xlttlib_FontInfoType;

/**************************************************************************/
/*                                                                        */
/* Exported function prototypes                                           */
/*                                                                        */
/**************************************************************************/

/**************************************************************************
   Before xlttlib can be used, the following function must be called to
   initialize the library and pass it the memory access functions.
*/

extern boolean xlttlib_Initialize
(
  xlttlib_MemAccessType memAccess
);

/**************************************************************************
    The following function must be called for every font that the client
    wants to use. This function returns useful information back to the
    client in the form of a xlttlib_FontInfoType data structure.
*/

extern boolean xlttlib_RegisterFont
(
  xlttlib_FontAccessType  fontAccess,
  xlttlib_FontInfoType   *fontInfoPtr
);

/**************************************************************************/

extern boolean xlttlib_DownloadEntireFont
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName    /* must be NULL terminated */ 
);

/**************************************************************************/

extern boolean xlttlib_DownloadHeader
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName    /* must be NULL terminated */ 
);

/**************************************************************************/

extern boolean xlttlib_DownloadCharacters
(
  HP_StreamHandleType    pStream,
  xlttlib_FontHandleType fontHandle,
  HP_pUByte              fontName,
  HP_UInt16              startCode,
  HP_UInt16              endCode
);

/**************************************************************************
    PCL-XL expects the application to send widths for each character to be
    printed. The following function can be used to get scaled character
    widths for a range of characters from the specified font at a specified
    charSize (user units). Note that widths will be linearly scaled. That
    is if the character's size increases by a factor of 2 so will it's width.
    The widthsArray must be allocated before this function is called and must
    be at least large enough to store ((endCode-startCode)+1) entries. For
    proportional DBCS fonts, the character width may be different if the font
    is vertical-rotated, so be sure to set the verticalRotated flag whenever
    the characters will be printed in a vertical rotated fashion. Also,
    vertical substitute glyphs may have different metrics from their
    horizontal counterparts, so be sure to set the verticalSubstitution flag
    if the characters will be printed in a vertical substituted fashion.
*/

extern boolean xlttlib_GetLinearScaledCharWidths
(
  xlttlib_FontHandleType  fontHandle,
  HP_Real32               charSize, 
  boolean                 verticalRotation,
  boolean                 verticalSubstitution,
  HP_UInt16               startCode,
  HP_UInt16               endCode,
  HP_pSInt16              widthsPtr /* preallocated, size=(end-start)+1 */
);

/**************************************************************************
    The following function should be called when the client has finished
    using a registered font. This allows xlttlib to free any memory
    used to process the font. Calling this function WILL NOT delete it
    from the printer. To delete the font from the printer, the client
    must call the jetlib function, HP_RemoveFont_1.
*/

extern boolean xlttlib_UnregisterFont
(
  xlttlib_FontHandleType fontHandle
);

/**************************************************************************
    The following function should be called when the client has finished
    using xlttlib. This allows xlttlib to free any memory it has allocated.
    Calling this function WILL NOT delete fonts from the printer. To delete
    fonts from the printer, the client must call the jetlib function,
    HP_RemoveFont_1 for each font.
*/

extern boolean xlttlib_ShutDown
(
  void
);

/**************************************************************************
    If any of the xlttlib routines return failure, the client should call
    the following function, which will tell the client what the error was.
*/

extern xlttlib_ErrorType xlttlib_GetError
(
  void
);

/**************************************************************************/

#endif

